import logging
import time

from naludaq.backend import ConnectionManager
from naludaq.controllers import get_board_controller
from naludaq.tools.data_collector.default import DataCollector

logger = logging.getLogger("naludaq.data_collector_udc16")

DEFAULT_TRIGGER_INTERVAL_LIMIT = 0.001


class Udc16DataCollector(DataCollector):
    """Data collector for the UDC16.

    This subclass is necessary for a few reasons:
    - Starting a readout is different for the UDC16
    - The UDC16 has no concept of a read window
    - The UDC16 needs to be reset when an event times out
    """

    def __init__(self, board, *args, **kwargs):
        super().__init__(board, *args, **kwargs)

        # UDC16/UPAC96 have an issue where the firmware locks up if the triggers
        # are sent too close together
        self.trigger_interval_limit = board.params.get(
            "trigger_interval_limit", DEFAULT_TRIGGER_INTERVAL_LIMIT
        )

    def _start_readout(self):
        logger.info("Starting readout")
        bc = get_board_controller(self.board)
        bc.clear_buffer()
        time.sleep(0.2)
        self._clear_connection_buffer()
        if self.board.model in ["upac96"]:
            bc.set_continuous_mode(True)
        bc.start_readout("software")

    def _restart_readout(self):
        self._stop_readout()
        time.sleep(0.1)
        bc = get_board_controller(self.board)
        bc.digital_reset()
        bc.digital_reset()
        time.sleep(0.1)
        self._start_readout()

    def _setup_readout(self):
        """Does nothing since UDC16 has nothing to configure"""

    def _event_failed(self):
        self._restart_readout()
        super()._event_failed()

    def _clear_connection_buffer(self):
        if self._board.using_new_backend:
            ConnectionManager(self.board).device.clear_buffers()
        else:
            self._board.connection.reset_input_buffer()



"""Package creator is the second part in a multithread double buffer.
The packager slices teh raw data from the serial reader into packages.

OVERVIEW:
Checks the applications serial buffer for data.
Reads all data in the buffer and splits the data into packages.

Return package:
{'created_at': time.time(),
 'length': length of data,
 raw_data': bytesarray()
}

"""

import logging
import time
from collections import deque
from threading import Thread

logger = logging.getLogger("naludaq.packager")


##############################################################################
class OldPackagerLight(Thread):
    """Drop in replacement for the Packager but without the zmq stuff.

    Significantly slower, beware.
    """

    def __init__(self, input_buffer, output_buffer, stop_word, frequency):
        super().__init__()
        self.daemon = True
        self.input_buffer = input_buffer
        self.output_buffer = output_buffer
        self.stop_word = stop_word
        self.frequency = frequency
        self._reset_pkg_num = False
        self.running = True
        self._pkg_num = 0
        self.answer_size = 8

    @property
    def frequency(self):
        """Get/Set the frequency this thread polls the serial port.

        Frequency in Hz (times per second).
        """

        return self._frequency

    @frequency.setter
    def frequency(self, freq):
        if not isinstance(freq, (int, float)):
            raise TypeError("Frequency must be an integer or float")
        if freq <= 0:
            raise ValueError("Frequency must be positive")

        self._frequency = freq
        self._interval = 1 / freq

    @property
    def stop_word(self):
        """Get/Set the stop word, the end word of a package.

        Args:
            stop_word (bytes): Specific word ending an event in input_buffer.
        Returns:
            Returns the current stop word as bytes.
        """
        return self._stop_word

    @stop_word.setter
    def stop_word(self, stop_word):
        if isinstance(stop_word, bytes):
            self._stop_word = stop_word
        else:
            raise TypeError("stop_word must be bytes")

    def reset_pkg_num(self):
        """Resets the event numbering.

        Raises the reset flag and let the loop reset the evt_num
        """
        self._reset_pkg_num = True

    def run(self):
        logger.debug("Trying to start packager.")
        self._package_worker()

    def stop(self):
        self.running = False

    def _slice_buffer(self, internal_buffer):
        """Slice the internal buffer at the stopword.

        Takes an internal buffer as a bytearray and cuts it at the stopword.
        The output is a list of packages.
        Once sliced the last part will be the first part of the next batch.

        Args:
            internal_buffer (bytearray): raw data to slice.

        Returns:
            output: a list of packages
            rest: whatever is after the last stopword.
        """

        packages = internal_buffer.split(self._stop_word)

        output = [
            {
                "created_at": time.time(),
                "length": len(x + self._stop_word),
                "pkg_num": self._pkg_num + idx,
                "rawdata": x + self._stop_word,
            }
            for idx, x in enumerate(packages[:-1])
        ]

        self._pkg_num += len(packages) - 1
        rest = packages[-1]
        del packages

        return output, rest

    def _package_worker(self):
        """Worker takes raw serial string and package into packages.

        It looks for the end of the package b'face'.
        Ideally it would find the header of the package, then read until the end.
        It adds a timestamp on when the package is arrived for sorting purposes.

        Appends to package_buffer {'package_id': timestamp, 'length': length, rawdata':package}.
        """
        internal_buffer = bytearray()
        try:
            while self.running or self.input_buffer:
                frametime = time.perf_counter()

                if self.input_buffer:
                    try:
                        internal_buffer.extend(self.input_buffer.pop())
                    except:
                        logger.error("can't pop!")

                if internal_buffer:
                    output, rest = self._slice_buffer(internal_buffer)
                    internal_buffer = bytearray()
                    internal_buffer.extend(rest)
                    for package in output:
                        package["event_num"] = package["pkg_num"]
                        self.output_buffer.append(package)

                if self._reset_pkg_num:
                    self._pkg_num = 0
                    self._reset_pkg_num = False

                try:
                    time.sleep(frametime + self._interval - time.perf_counter())
                except ValueError:
                    # LOGGER.debug("Package worker is %ss behind",
                    #              (frametime + self._interval - time.perf_counter()))
                    pass
        except KeyboardInterrupt:
            pass


##############################################################################
class PackagerLight(Thread):
    """Drop in replacement for the Packager but without the zmq stuff.

    Significantly slower, beware.
    """

    def __init__(
        self,
        input_buffer: deque,
        output_buffer: deque,
        output_answers: deque,
        stop_word: bytes,
        frequency: int,
    ):
        """Prepares a thread (not open) which polls the input buffer and pops all data in it
        into a internal buffer. Depending on frequency, it will slice the internal
        buffer once it reaches a stop word, and store the sliced data into an
        output_buffer for parsing. If there are answers within the data, it will separate it
        and store it in output_answers.

        Args:
            input_buffer (deque): Incoming stream from serial reader, raw bytes
            output_raw_buffer (deque): Stores raw sliced data if defined
            output_buffer (deque): Stores dict'd raw sliced data + info about pkg
            output_answers (deque): Data recieved intermittent to readout
            stop_word (bytes): Stop word to slice on
            frequency (int): How quick to poll the input buffer and slice
        """
        super().__init__()
        self.daemon = True
        self.input_buffer = input_buffer
        self.output_buffer = output_buffer
        self.answers = output_answers
        self.stop_word = stop_word
        self.frequency = frequency
        self._reset_pkg_num = False
        self.running = True
        self._pkg_num = 0
        self.answer_size = 8

    @property
    def frequency(self):
        """Get/Set the frequency this thread polls the serial port.

        Frequency in Hz (times per second).
        """

        return self._frequency

    @frequency.setter
    def frequency(self, freq):
        if not isinstance(freq, (int, float)):
            raise TypeError("Frequency must be an integer or float")
        if freq <= 0:
            raise ValueError("Frequency must be positive")

        self._frequency = freq
        self._interval = 1 / freq

    @property
    def stop_word(self):
        """Get/Set the stop word, the end word of a package.

        Args:
            stop_word (bytes): Specific word ending an event in input_buffer.
        Returns:
            Returns the current stop word as bytes.
        """
        return self._stop_word

    @stop_word.setter
    def stop_word(self, stop_word):
        if isinstance(stop_word, bytes):
            self._stop_word = stop_word
        else:
            raise TypeError("stop_word must be bytes")

    def reset_pkg_num(self):
        """Resets the event numbering.

        Raises the reset flag and let the loop reset the evt_num
        """
        self._reset_pkg_num = True

    def run(self):
        logger.debug("Trying to start packager.")
        self._package_worker()

    def stop(self):
        self.running = False

        return self._pkg_num

    def _is_answer_package(self, package: bytearray) -> bool:
        """Checks if a package is an answer package based on length.

        Args:
            package (bytearray): the raw package.

        Returns:
            True if the package is an answer, False otherwise.
        """
        return len(package) + len(self.stop_word) <= self.answer_size

    def _is_data_package(self, package: bytearray) -> bool:
        """Checks if a package is a data (event) package based on length.

        Args:
            package (bytearray): the raw package.

        Returns:
            True if the package is data, False otherwise.
        """
        return len(package) + len(self.stop_word) > self.answer_size

    def _create_answer_package(self, rawdata: bytearray) -> dict:
        """Creates an answer package from rawdata.

        Args:
            rawdata (bytearray): the raw data.

        Returns:
            The answer package as a dict.
        """
        return {
            "created_at": time.time(),
            "length": len(rawdata),
            "rawdata": rawdata,
        }

    def _create_data_package(self, rawdata: bytearray, pkg_num) -> dict:
        """Creates a data package from rawdata.

        Args:
            rawdata (bytearray): the raw data.

        Returns:
            The data package as a dict.
        """
        return {
            "created_at": time.time(),
            "length": len(rawdata),
            "pkg_num": pkg_num,
            "rawdata": rawdata,
        }

    def _slice_buffer(self, internal_buffer):
        """Slice the internal buffer at the stopword.

        Takes an internal buffer as a bytearray and cuts it at the stopword.
        The output is a list of packages.
        Once sliced the last part will be the first part of the next batch.

        Args:
            internal_buffer (bytearray): raw data to slice.

        Returns:
            A tuple of (output, answers, rest):
                - output: a list of event packages
                - answers: a list of answer packages
                - rest: whatever is after the last stopword.
        """
        packages = internal_buffer.split(self._stop_word)

        output = [
            self._create_data_package(x + self._stop_word, self._pkg_num + idx)
            for idx, x in enumerate(packages[:-1])
            if self._is_data_package(x)
        ]
        answers = [
            self._create_answer_package(x + self._stop_word)
            for x in packages[:-1]
            if self._is_answer_package(x)
        ]

        self._pkg_num += len(packages) - 1
        rest = packages[-1]
        del packages

        return output, answers, rest

    def _route_serial_data(self, buff: bytearray):
        """Splits and routes serial data into the appropriate
        output buffers.

        Args:
            buff (bytearray): the serial input buffer

        Returns:
            Any leftover serial data following the last stop word.
        """
        output, answers, rest = self._slice_buffer(buff)
        for package in output:
            package["event_num"] = package["pkg_num"]
            self.output_buffer.append(package)
        for package in answers:
            self.answers.put(package)

        return rest

    def _package_worker(self):
        """Worker takes raw serial string and package into packages.

        It looks for the end of the package b'face'.
        Ideally it would find the header of the package, then read until the end.
        It adds a timestamp on when the package is arrived for sorting purposes.

        Appends to package_buffer {'package_id': timestamp, 'length': length, rawdata':package}.
        """
        internal_buffer = bytearray()
        try:
            while self.running:
                frametime = time.perf_counter()

                for _ in range(len(self.input_buffer)):
                    try:
                        internal_buffer.extend(self.input_buffer.pop())
                    except:
                        logger.error("can't pop!")

                if internal_buffer:
                    rest = self._route_serial_data(internal_buffer)
                    internal_buffer = bytearray(rest)

                if self._reset_pkg_num:
                    self._pkg_num = 0
                    self._reset_pkg_num = False

                try:
                    time.sleep(frametime + self._interval - time.perf_counter())
                except ValueError:
                    pass
        except KeyboardInterrupt:
            pass



import logging
import time
from datetime import datetime
from os import PathLike
from os.path import exists

import h5py
import numpy as np

from naludaq.board import Board
from naludaq.tools.dac_sweep.dac_sweep_controller import DACSweepController
from naludaq.tools.lookup_table import LookupTable

LOGGER = logging.getLogger(__name__)


class LookupTableGenerator:
    """
    Class for generating/continuing a lookup table for ADC2mV

    Example:
        LookupTableGenerator.initialize_lookuptable('file', board)
        with LookupTable('file', 'r+') as lookup_table:
            ltg = LookupTableGenerator(lookup_table, board)
            ltg.dac_sweep(
                min_counts=1000,
                max_counts=3000,
                step_size=1000,
                events_per_datapt=1,
                num_warmup_events=1
                )
            ltg.interpolate_data()
    """

    def __init__(
        self,
        lookup_table: LookupTable,
        board: Board,
    ):
        self.lookup_table = lookup_table
        self.board = board
        self._dac_sweep_ctrl = None

    def dac_sweep(
        self,
        min_counts: int = 0,
        max_counts: int = None,
        step_size: int = 50,
        events_per_datapt: int = 3,
        num_warmup_events: int = 3,
        channels: list = None,
        mode: str = "overwrite",
    ):
        """
        Perform a DAC sweep to collect data for the lookup table, follows
        np.linspace when determining range

        Args:
            min_counts (int): Min count to start the sweep
            max_counts (int): Max count that the sweep will end at
            step_size (int): Step size to increment by
            events_per_datapt (int): Number of events to capture per DAC Value
            num_warmup_events (int): Number of events to capture/discard before capturing data
            mode (str): 'overwrite' or 'average' existing data points
        """
        validate_table_params_or_raise(self.lookup_table, self.board)

        if max_counts == None:
            max_counts = self.board.params["ext_dac"]["max_counts"]
        if channels is None:
            channels = list(range(self.board.channels))
        self._validate_sweep_params(step_size, max_counts, min_counts, mode)

        self._dac_sweep_ctrl = DACSweepController(self.board)
        self._dac_sweep_ctrl._backup_settings()
        try:
            for dac_val in range(min_counts, max_counts + 1, step_size):
                LOGGER.info("Running dac value %d", dac_val)
                dac_val = int(dac_val)
                self.run_dac_value(
                    dac_val, channels, events_per_datapt, num_warmup_events, mode
                )
        finally:
            self._dac_sweep_ctrl._restore_settings()

    def run_dac_value(
        self,
        dac_val,
        channels,
        events_per_datapt: int = 3,
        num_warmup_events: int = 3,
        mode: str = "overwrite",
    ):
        """Runs a dac capture at the given dac value and stores the
        data into the captured_data table then updates the lookup table.

        Args:
            dac_data (list): entry of data captured using the dac sweep controller
            dac_val (int): dac value used to capture the data
            channels (list): channels to write
            events_per_datapt (int): Number of events to capture for pedestals
            num_warmup_events (int): Number of warmup events for pedestals
            mode (str): 'overwrite' or 'average' how to handle repeat data
        """
        if self._dac_sweep_ctrl is None:
            self._dac_sweep_ctrl = DACSweepController(self.board)

        dac_pedestals = self._dac_sweep_ctrl._run_dac_value(
            dac_val=dac_val,
            channels=channels,
            num_captures=events_per_datapt,
            num_warmup_events=num_warmup_events,
        )
        dac_data = dac_pedestals["data"]
        LOGGER.debug("Dac value: %d total events: %d", dac_val, len(dac_data))
        self.write_dac_data(dac_data, dac_val, channels, mode)

    def write_dac_data(self, dac_data, dac_val, channels, mode: str = "overwrite"):
        """Writes an entry of the data generated by the dac sweep controller
        to captured_data table, converting dac_val to mV. Updates the times
        captured table accordingly, and updates the lookup table to the
        captured_data table.

        Args:
            dac_data (list): pedestals data captured using the dac sweep controller
            dac_val (int): dac value used to capture the data
            channels (list): channels to write
            mode (str): 'overwrite' or 'average' how to handle repeat data
        """
        max_mv = self.board.params["ext_dac"]["max_mv"]
        mV = max_mv * dac_val / self.board.params["ext_dac"]["max_counts"]
        LOGGER.info("Writing data to file for Dac val: %d", dac_val)
        for channel in channels:
            peds = np.array(dac_data[channel]).flatten()
            for sample_num, adc_count in enumerate(peds):
                capture_idx = (int(round(adc_count)), channel, sample_num)
                self.write_captured_data(capture_idx, mV, mode)
        # Write captured data to lookup table dataset
        self.update_lookup_table()

    def write_captured_data(self, capture_idx, mV: int, mode: str = "overwrite"):
        """Writes captured data to captured_data table and
        updates the times captured table accordingly

        Args:
            capture_idx (tuple or arr): Index of captured data table to write
            to in form: (adc_count (int), channel (int), sample_num (int))
            mV (float): mV value to set
            mode (str): 'overwrite' or 'average' how to handle repeat data
        """
        if not mode in ["overwrite", "average"]:
            raise ValueError("Invalid mode, 'overwrite' or 'average' is accepted")
        try:
            if mode == "overwrite" or self.times_captured[capture_idx] == 0:
                self._overwrite_data(capture_idx, mV)
            if mode == "average":
                self._average_data(capture_idx, mV)
        except KeyboardInterrupt:
            self.lookup_table.captured_data[capture_idx] = np.nan
            self.lookup_table.times_captured[capture_idx] = 0
            raise

    def write_event(self, event: dict, mV: float, mode: str = "overwrite"):
        """Writes captured events to captured_data table and
        updates the times captured table accordingly

        Args:
            event (dict): Parsed event
            mV (float): mV value to set
            mode (str): 'overwrite' or 'average' how to handle repeat data
        """
        if not all(x in event.keys() for x in ["data", "window_labels"]):
            raise KeyError("Event does not contain 'data' and/or 'window_labels'")
        samples = self.board.params["samples"]
        for channel in range(self.board.channels):
            for idx, window in enumerate(event["window_labels"][channel]):
                for relative_sample in range(samples):
                    sample_num = window * samples + relative_sample
                    adc_count = int(
                        event["data"][channel][relative_sample + idx * samples]
                    )
                    capture_idx = (adc_count, channel, sample_num)
                    self.write_captured_data(capture_idx, mV, mode)
        self.update_lookup_table()

    def interpolate_data(self):
        """Fills empty entries in from captured data, and stores it
        to lookup table with linear interpolation. Left and right
        ends of the dac entries are filled with min and max values
        for dac value respectively.
        """
        channels = self.board.params["channels"]
        for channel in range(channels):
            self.interpolate_channel_data(channel)

    def interpolate_channel_data(self, channel: int):
        """Fills empty entries in from captured data for a channel, and stores
        it to lookup table with linear interpolation.
        Left and right ends of the dac entries are filled with
        min and max values for dac value respectively.
        """
        validate_table_params_or_raise(self.lookup_table, self.board)
        num_samples = self.board.params["samples"] * self.board.params["windows"]
        data_width = self.board.params["ext_dac"]["max_counts"] + 1
        channel_slice = np.array(self.lookup_table[:, channel, :])
        channel_slice[0][:] = np.nanmin(channel_slice, axis=0)
        channel_slice[-1][:] = np.nanmax(channel_slice, axis=0)
        # Get indices of captured data
        data_indices = np.argwhere(
            ~np.isnan(channel_slice.transpose().ravel())
        ).flatten()
        interpolated_data = np.interp(
            range(data_width * num_samples),
            data_indices,
            channel_slice.transpose().ravel()[data_indices],
        )
        self.lookup_table[:, channel, :] = interpolated_data.reshape(
            (num_samples, data_width)
        ).transpose()

    def update_lookup_table(self):
        """Updates lookup table to captured_data table"""
        self.lookup_table.lookup[...] = self.lookup_table.captured_data

    def _overwrite_data(self, capture_idx, mV: int):
        """Overwrites captured data to captured_data table and
        updates the times captured table accordingly to 1
        """
        self.lookup_table.captured_data[capture_idx] = mV
        self.lookup_table.times_captured[capture_idx] = 1

    def _average_data(self, capture_idx, mV: int):
        """Averages captured data to captured_data table and
        updates the times captured table accordingly
        """
        cur_val = self.lookup_table.captured_data[capture_idx]
        times_cap = self.lookup_table.times_captured[capture_idx]
        avg = (cur_val * times_cap + mV) / (times_cap + 1)
        self.lookup_table.captured_data[capture_idx] = avg
        self.lookup_table.times_captured[capture_idx] += 1

    def _process_avg_ped(self, dac_data, channel):
        """Converts dac_data to pedestals"""
        ped_list = list()
        for capture_num in range(len(dac_data[0])):
            ped_data = [event[capture_num]["data"][channel] for event in dac_data]
            ped_data = np.array(ped_data).flatten().astype(int)
            ped_list.append(ped_data)
        avg_ped = np.array(ped_list).mean(axis=0).astype(int)

        return avg_ped

    @staticmethod
    def initialize_lookuptable(file_name: str, board: Board):
        """Writes a file and initializes three tables for the given board with two formats:
        lookup_table, captured_data:
        [ADC Counts][Channels][Windows * Samples] = mV (float)
        times_captured:
        [ADC Counts][Channels][Windows * Samples] = Num Captures (int)
        Initialized to np.nan for mV Values, and 0 for Num Captures
        Does not create a new lookup table if the file already exists

            Args:
                filename (str): Filename/path to store lookup table
                board (Board): Board object to initialize the lookup table's dimensions
        """

        if exists(file_name):
            raise FileExistsError("Lookup table already exists")

        with h5py.File(file_name, "w") as h5_file:
            h5_file.attrs.create("created_at", time.time())
            settings = {
                "model": board.params["model"],
                "params": board.params,
                "registers": board.registers,
            }
            settings_group = h5_file.create_group("settings")
            _store_attributes(settings_group, settings)
            windows = board.params["windows"]
            samples = board.params["samples"]
            data_width = 2 ** board.params["resolution"]
            channels = board.params["channels"]

            h5_file.create_dataset(
                name="captured_data",
                shape=(data_width, channels, windows * samples),
                compression="gzip",
                compression_opts=3,
                fillvalue=np.nan,
                dtype=np.float32,
            )
            h5_file.create_dataset(
                name="lookup_table",
                shape=(data_width, channels, windows * samples),
                compression="gzip",
                compression_opts=3,
                fillvalue=np.nan,
                dtype=np.float32,
            )
            h5_file.create_dataset(
                name="times_captured",
                shape=(data_width, channels, windows * samples),
                compression="gzip",
                compression_opts=3,
                fillvalue=0,
                dtype=np.int16,
            )

    def _validate_sweep_params(self, step_size, max_counts, min_counts, mode):
        if step_size < 1:
            raise ValueError("step_size must be > 0")
        if max_counts < min_counts:
            raise ValueError("Invalid min/max counts")
        if step_size > max_counts:
            raise ValueError("Invalid step_size value")
        if not mode in ["overwrite", "average"]:
            raise ValueError("Invalid mode, 'overwrite' or 'average' is accepted")


def validate_table_params_or_raise(lookup_table: LookupTable, board: Board) -> bool:
    """Validates the table dimensions from the lookuptable file, to the
    given board object
    """

    keys = [
        "lookup_table",
        "times_captured",
        "captured_data",
        "settings",
    ]
    h5_file = lookup_table._h5_file
    for key in keys:
        if key not in h5_file.keys():
            raise KeyError(f"Dataset doesn't exist {key}")

    if h5_file["settings"].attrs["model"] != board.model:
        raise ValueError("Given board model does not match the lookup table")
    h5_lookup_table = h5_file["lookup_table"]
    if (board.params["ext_dac"]["max_counts"] + 1) != h5_lookup_table.shape[0]:
        raise ValueError("Data width parameter does not match the given board")
    if board.params["channels"] != h5_lookup_table.shape[1]:
        raise ValueError("Channels parameter does not match the given board")
    if board.params["samples"] * board.params["windows"] != h5_lookup_table.shape[2]:
        raise ValueError(
            "Sample and/or windows parameters do not match the given board"
        )


def _store_attributes(group: h5py.Group, attrs: dict, recursive=True):
    """Adds a dict of attributes to a h5py group.

    Skips entries that have 'None' value.

    Args:
        group (Group): the group to add the attributes to
        attrs (dict): a dict of {name: value}
        recursive (bool): whether to create new groups for
            entries with dict values
    """
    for k, v in attrs.items():
        k = str(k)
        if v is None:  # HDF5 doesn't support None value
            continue
        elif isinstance(v, dict) and recursive:
            new_group = group.create_group(k, track_order=True)
            _store_attributes(new_group, v)
        else:
            v = _convert_for_storing(v)
            group.attrs.create(k, v)


def _convert_for_storing(value):
    """Converts a value to a type that can be stored in HDF5

    datetime -> float
    PathLike -> str
    bytes -> np.void

    Args:
        value: the value to convert

    Returns:
        A tuple of (converted value, new dtype)
    """
    if isinstance(value, datetime):
        value = value.timestamp()
    elif isinstance(value, PathLike):
        value = str(value)
    elif isinstance(value, bytes):
        value = np.void(value)
    return value



import logging

from naludaq.controllers import get_board_controller
from naludaq.helpers.exceptions import IterationError, RegisterNameError
from naludaq.helpers.helper_functions import (
    find_missing_channels,
    group_channels_by_chip,
)

from .udc16 import PedestalsGeneratorUdc16

LOGGER = logging.getLogger("naludaq.pedestals_generator_upac96")


def disable_trigger_monitor_signal(func):
    """Decorator to disable the trigger monitor signal during the execution of a function.

    Args:
        func (function): function to decorate

    Returns:
        function: decorated function
    """

    def wrapper(self, *args, **kwargs):
        try:
            previous = self.board.registers["control_registers"][
                "trigger_monitor_disable"
            ]["value"]
        except KeyError:
            raise RegisterNameError("trigger_monitor_disable")
        bc = get_board_controller(self.board)
        bc.set_trigger_monitoring_disabled(disabled=True)
        try:
            result = func(self, *args, **kwargs)
        finally:
            bc.set_trigger_monitoring_disabled(previous)
        return result

    return wrapper


class PedestalsGeneratorUpac96(PedestalsGeneratorUdc16):
    """Pedestals generator for UPAC96."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # last event, regardless of whether it's valid or not
        self._last_event_channels = []

    @disable_trigger_monitor_signal
    def _capture_data_for_pedestals(self) -> list[list[dict]]:
        return super()._capture_data_for_pedestals()

    def _validate_event(self, event):
        """Check if the event has a data field, which means it's parsed"""
        self._last_event_channels = []
        if "data" not in event:
            LOGGER.warning("Got an invalid event")
            return False
        chans_with_data = [i for i, x in enumerate(event.get("data", [])) if len(x) > 0]
        self._last_event_channels = chans_with_data
        is_superset = set(chans_with_data).issuperset(self._channels)
        if not is_superset:
            LOGGER.warning(
                "Got a parseable event, but the channels are incorrect: %s",
                chans_with_data,
            )
        return is_superset

    def _exc_str(self, e: Exception) -> str:
        """Override since UPAC96 can fail on a per-chip basis"""
        missing_channels = find_missing_channels(
            self._channels, self._last_event_channels
        )
        # IterationError means we're getting valid data but it's getting filtered
        # out by the constraints in `validate_event()`. Got a custom message for that.
        if not isinstance(e, IterationError) or len(missing_channels) == 0:
            return super()._exc_str(e)

        channels_per_chip = self.board.params.get("channels_per_chip", 16)
        missing_channels = group_channels_by_chip(missing_channels, channels_per_chip)

        chip_parts = []
        for chip, missing_in_chip in missing_channels.items():
            if len(missing_in_chip) == 0:
                continue
            msg = f"- Chip {chip}: all channels"
            if len(missing_in_chip) < channels_per_chip:
                chan_repr = ", ".join(map(str, missing_in_chip))
                msg = f"- Chip {chip}: channel(s) {chan_repr}"
            chip_parts.append(msg)

        missing_formatted = "\n".join(chip_parts)
        return (
            "Failed to generate pedestals calibration because the following chips/channels "
            "cannot be read from the hardware:\n"
            f"{missing_formatted}\n\n"
            "Please try again and power cycle the board if the issue persists. To instead ignore this error, "
            "deselect the affected channel(s) and try again."
        )



"""DebugDaq
=======

A DAQ used to collect more information about the incoming data.

"""
from collections import deque
from logging import getLogger
from queue import Queue

from naludaq.backend.exceptions import BackendError
from naludaq.daq.workers import get_usb_reader
from naludaq.daq.workers.packager import DebugPackager, get_packager
from naludaq.daq.workers.worker_parser import ParserWorkerLight
from naludaq.parsers import get_parser

LOGGER = getLogger(__name__)

HDSOC_PACKAGER = ["hdsocv1", "hdsocv1_evalr1", "hdsocv1_evalr2"]


class DebugDaq:
    def __init__(self, board, store_raw_data=False, custom_parser=None):
        """Creates a DAQ used to collect data readouts from the given board.

        Args:
            board (naludaq.board): Board object to collect data from
            store_raw_data (bool, optional): If True, the raw bytearray
                collected from the board will be saved to a buffer.
                Defaults to False.
            custom_parser (parser object, optional): A user provided parser which
                will be used to parse raw data. Requires a "parse" attribute.
                Defaults to None.
        """
        if board.using_new_backend:
            raise BackendError("DAQs are incompatible with the new backend")
        self.board = board
        self.serial_to_packager_buffer = deque()
        self.packager_to_parser_buffer = deque()
        self.parser_error_buffer = deque()
        self.intermittent_answers_buffer = Queue()
        self._output_buffer = deque()
        self._raw_output_buffer = deque()
        self._acquisition_name = None
        self._custom_parser = None
        self._store_raw_data = store_raw_data
        self.pedestals = None
        self.workers = list()
        self._custom_parser = custom_parser

    @property
    def output_buffer(self) -> deque:
        """The buffer that stores events from the readout."""
        return self._output_buffer

    @output_buffer.setter
    def output_buffer(self, buffer):
        self._output_buffer = buffer

    @property
    def raw_output_buffer(self) -> deque:
        """The buffer that stores raw data from events. Each element added is a dict containing an unparsed 'rawdata' bytearray."""
        return self._raw_output_buffer

    @raw_output_buffer.setter
    def raw_output_buffer(self, buffer):
        self._raw_output_buffer = buffer

    @property
    def raw_output_stream(self) -> bytearray:
        """
        Takes the raw_output_buffer property and joins all bytearrays into a
        single bytearray, giving one continuous stream.
        """
        return b"".join([package["rawdata"] for package in self.raw_output_buffer])

    @property
    def custom_parser(self):
        """A custom parser to use for parsing raw events. Must be set _before_ starting capture.

        Raises:
            AttributeError if set to an object that does not have a `parse` attribute.
        """
        return self._custom_parser

    @custom_parser.setter
    def custom_parser(self, parser):
        if parser is None or self.validate_parser(parser):
            self._custom_parser = parser
        else:
            raise AttributeError(
                "The given custom parser does not have attribute 'parse'."
            )

    @property
    def acquisition_name(self):
        """Get/Set the Acquisition name."""
        return self._acquisition_name

    @acquisition_name.setter
    def acquisition_name(self, name):
        self._acquisition_name = name

    @property
    def board(self):
        """Change the board controller used.

        Useful if the board model is changed, then update the board controller.
        This can also be used to restart the daq since all Threads and Processes are stopped
        before changing the board controller.

        Args:
            board_controller: New board controller to use.
        """
        return self._board

    @board.setter
    def board(self, board):
        LOGGER.debug("DAQ: changing board to: %s", board.params["model"])

        self._board = board

    def start_capture(self):
        """Starts workers that listen for readout data. Parsed events are stored in the
        `output_buffer` attribute, and raw events are stored in the `raw_output_buffer`
        attribute.

        If called while still capturing, the workers are restarted.
        """
        if self.workers:
            self.stop_capture()

        serial_reader = self._get_serial_reader()
        packager_worker = self._get_packager_worker()
        parser_worker = self._get_parser_worker()

        self.workers.append(serial_reader)
        self.workers.append(packager_worker)
        self.workers.append(parser_worker)

        for worker in self.workers:
            worker.start()

    def _get_serial_reader(self):
        return get_usb_reader(self.board, self.serial_to_packager_buffer)

    def _get_parser_worker(self):
        if self._custom_parser is None:
            parser = get_parser(self.board.params)
        else:
            parser = self.custom_parser

        pw = ParserWorkerLight(
            parser,
            self.packager_to_parser_buffer,
            self.output_buffer,
            self.parser_error_buffer,
        )
        return pw

    def _get_packager_worker(self):

        if self._store_raw_data:
            raw_buffer = self._raw_output_buffer
        else:
            raw_buffer = None

        if self.board.params["model"] in HDSOC_PACKAGER:
            dp = get_packager(
                self.board,
                self.serial_to_packager_buffer,
                # raw_buffer,
                self.packager_to_parser_buffer,
                self.intermittent_answers_buffer,
                self.board.params["stop_word"],
                1000,
            )
        else:
            dp = DebugPackager(
                # self.board,
                self.serial_to_packager_buffer,
                raw_buffer,
                self.packager_to_parser_buffer,
                self.intermittent_answers_buffer,
                self.board.params["stop_word"],
                1000,
            )

        return dp

    def stop_capture(self):
        """Stops all workers currently running."""
        while self.workers:
            worker = self.workers.pop()
            worker.stop()
            worker.join()

    def stop_workers(self):
        """Does nothing, but makes the class look like a LightDaq."""
        pass

    def switch_output_buffer(self, new_buffer, pedestals=None):
        """Switching the output buffer, the daq will finish the last batch
        before switching.

        Args:
            new_buffer: Acquisition or deque to append to.
            pedestals: Pedestals to use with next acquisition.
        """
        self.output_buffer = new_buffer
        self.pedestals = pedestals

        if "name" in new_buffer.__dir__():
            self.acquisition_name = new_buffer.name

    def validate_parser(self, parser):
        """Checks whether the given custom parser can be used for parsing events.

        Returns:
            True if the object has a "parse" attribute, or False otherwise.
        """
        return hasattr(parser, "parse")



import time
from http import HTTPStatus

from naludaq.backend.client import HttpClient
from naludaq.backend.exceptions import DeviceError, HttpError
from naludaq.backend.managers.base import Manager
from naludaq.backend.models.device import (
    AvailableD2xxDevice,
    AvailableD3xxDevice,
    AvailableSerialDevice,
    D2xxDevice,
    D3xxDevice,
    Device,
    DeviceType,
    SerialDevice,
    UdpDevice,
)


class ConnectionManager(Manager):
    def __init__(self, board):
        """Utility for controlling the board connection on the backend.

        Args:
            context (Context): context used to communicate with the backend.
        """
        super().__init__(board)

    @property
    def is_connected(self) -> bool:
        return self.device is not None

    @property
    def is_uart_based(self) -> bool:
        """Check if the connected device is UART-based (serial & D2XX).
        Will return `False` if there is no connection.
        """
        return hasattr(self.device, "baud_rate")

    @property
    def device(self) -> "Device | None":
        """Get a wrapper which can be used to configure the device connected to the backend.
        If no device is connected, `None` is returned.
        """
        try:
            type_name = self.context.client.get_json("/connection/info").get(
                "connection_type", None
            )
        except HttpError as e:
            if e.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
                raise DeviceError("Failed to retrieve device information")

        device_type = None
        if type_name is not None:
            device_type = DeviceType.from_string(type_name)
        if device_type == DeviceType.UDP:
            return UdpDevice(self.context)
        if device_type == DeviceType.SERIAL:
            return SerialDevice(self.context)
        if device_type == DeviceType.D2XX:
            return D2xxDevice(self.context)
        if device_type == DeviceType.D3XX:
            return D3xxDevice(self.context)
        return None

    def connect_udp(
        self,
        board_addr: tuple[str, int],
        receiver_addr: tuple[str, int],
        attempts: int = 5,
    ) -> UdpDevice:
        """Connect to the board through the backend using UDP.

        Any existing connections will be dropped prior to attempting to
        establish the new connection.

        Args:
            board_addr (tuple[str, int]): the board socket address
            receiver_addr (tuple[str, int]): the receiver socket address
            attempts (int): number of times to attempt connecting.

        Returns:
            UdpDevice: a handle to the device
        """
        board_addr = ":".join(map(str, board_addr))
        receiver_addr = ":".join(map(str, receiver_addr))
        arguments = {"board": board_addr, "receiver": receiver_addr}
        self._connect_with_attempts("udp", arguments, attempts)
        return UdpDevice(self.context)

    def connect_serial(
        self, port: str, baud_rate: int, attempts: int = 5
    ) -> SerialDevice:
        """Connect to the board through the backend using a serial port.

        Any existing connections will be dropped prior to attempting to
        establish the new connection.

        Args:
            port (str): COM port (Windows) or device file (Linux).
            baud_rate (int): serial connection baud rate.
            attempts (int): number of times to attempt connecting.
                Serial devices can randomly fail to open, so avoid
                setting this value too low.

        Returns:
            SerialDevice: a handle to the device
        """
        arguments = {"port": port, "baud_rate": baud_rate}
        self._connect_with_attempts("serial", arguments, attempts)
        return SerialDevice(self.context)

    def connect_d2xx(
        self, serial_number: str, baud_rate: int, attempts: int = 5
    ) -> D2xxDevice:
        """Connect to the board through the backend using D2XX.

        Any existing connections will be dropped prior to attempting to
        establish the new connection.

        Args:
            serial_number (str): D2XX device serial number.
            baud_rate (int): D2XX connection baud rate.
            attempts (int): number of times to attempt connecting.
                D2XX devices can randomly fail to open, so avoid
                setting this value too low.

        Returns:
            D2xxDevice: a handle to the device
        """
        arguments = {"serial_number": serial_number, "baud_rate": baud_rate}
        self._connect_with_attempts("d2xx", arguments, attempts)
        return D2xxDevice(self.context)

    def connect_d3xx(self, serial_number: str, attempts: int = 5) -> D2xxDevice:
        """Connect to the board through the backend using D3XX.

        Any existing connections will be dropped prior to attempting to
        establish the new connection.

        Args:
            serial_number (str): D3XX device serial number.
            attempts (int): number of times to attempt connecting.
                D3XX devices can randomly fail to open, so avoid
                setting this value too low.

        Returns:
            D3xxDevice: a handle to the device
        """
        arguments = {
            "serial_number": serial_number,
            "read_timeout_ms": 0,
            "write_timeout_ms": 0,
        }
        self._connect_with_attempts("d3xx", arguments, attempts)
        return D3xxDevice(self.context)

    def connect_from_info(self, info: dict) -> Device:
        """Connect to a device using an info dict.

        Args:
            info (dict): the info dict. Must contain "type" field and
                required arguments.

        Returns:
            Device: the connected device

        Raises:
            ValueError: if the info dict is invalid
        """
        type_name = info["type"].lower()
        if type_name == "udp":
            return self.connect_udp(info["board_addr"], info["receiver_addr"])
        if type_name == "serial":
            return self.connect_serial(info["port"], info["baud_rate"])
        if type_name == "d2xx":
            return self.connect_d2xx(info["serial_number"], info["baud_rate"])
        if type_name == "d3xx":
            return self.connect_d3xx(info["serial_number"])
        raise ValueError("Invalid connection info")

    def list_devices(
        self,
    ) -> "list[AvailableSerialDevice | AvailableD2xxDevice | AvailableD3xxDevice]":
        """Get a list of devices which are currently available to the backend.

        Returns:
            list[SerialDevice | D2xxDevice | D3xxDevice]: the list of devices.
        """
        return list_devices(self.context.client.address)

    def disconnect(self):
        """Attempts to drop any board connection the backend currently holds."""
        self.context.client.put("/connection/disconnect")

    def _connect_with_attempts(self, type_name: str, params: dict, attempts: int):
        """Open a device on the backend. Any existing connections will be dropped.

        Runs several attempts. This is mainly important for serial/D2XX, which
        can randomly fail to open for some reason.

        Args:
            type_name (str): connection type name (used for route).
            params (dict): connection arguments
            attempts (int): number of attempts.

        Raises:
            HttpError: if the device could not be opened.
        """
        ROUTE = f"/connection/{type_name}"
        for _ in range(attempts):
            self.disconnect()
            try:
                self.context.client.put(ROUTE, params=params)
            except HttpError:
                pass
            else:
                break
            time.sleep(0.1)
        else:
            raise DeviceError("Failed to connect to board")


def list_devices(
    backend_addr: tuple[str, int]
) -> "list[AvailableSerialDevice | AvailableD2xxDevice]":
    """Get a list of devices which are currently available to the backend.

    Returns:
        list[SerialDevice | D2xxDevice]: the list of devices.
    """
    client = HttpClient(backend_addr)
    response = client.get_json("/connection/list")
    result = []
    for device in response["devices"]:
        # UDP cannot be enumerated, so is not included in the response
        if "Serial" in device:
            result.append(AvailableSerialDevice(device["Serial"]["port"]))
        if "D2xx" in device:
            result.append(
                AvailableD2xxDevice(
                    serial_number=device["D2xx"]["serial_number"],
                    com_port=device["D2xx"].get("com_port", None),
                    index=device["D2xx"]["index"],
                    description=device["D2xx"]["description"],
                )
            )
        if "D3xx" in device:
            result.append(
                AvailableD3xxDevice(
                    serial_number=device["D3xx"]["serial_number"],
                    description=device["D3xx"]["description"],
                )
            )
    return result


def force_disconnect(backend_addr: tuple[str, int]):
    """Force the backend to disconnect any open connections.

    Useful for listing devices when the backend is hogging a board.

    Args:
        backend_addr (tuple[str, int]): address of the backend
    """
    client = HttpClient(backend_addr)
    client.put("/connection/disconnect")



def substitute_value(addr_value: int, register_value: int, register: dict):
    """Substitute the register value into the address value.

    Args:
        addr_value(int): The value of the address.
        register_value(int): The value of the register.
        register(dict): The register dict.

    Returns:
        The new value of the address.
    """
    bw = register["bitwidth"]
    bp = register["bitposition"]
    mask = (2**bw - 1) << bp
    return addr_value - (addr_value & mask) + ((register_value & (2**bw - 1)) << bp)


def normalize_address(addr: "int | str", width_addr: int) -> str:
    if not isinstance(addr, (int, str)):
        raise TypeError("Address must be a string or integer")
    if isinstance(addr, int):
        addr = f"{addr:X}"
    return addr.upper().zfill(width_addr)


def normalize_name(name: str) -> str:
    return name.casefold()


def full_to_partial_value(full_value: int, register: dict) -> int:
    """Strip out the value for a single register from a full 12-bit value."""
    position = register["bitposition"]
    width = register["bitwidth"]
    return (full_value >> position) & ((1 << width) - 1)


def partial_to_full_value(register: dict, value: int) -> int:
    """Convert the value for a single register into a 12-bit value, with the
    original value in the correct position.
    """
    position = register["bitposition"]
    width = register["bitwidth"]
    return (value & (1 << width) - 1) << position



"""

"""
from logging import getLogger

from naludaq.board import Board
from naludaq.tools.adc2mv.adc_linear_regression import ADCLinearReg
from naludaq.tools.dac_sweep.dac_sweep_controller import DACSweepController

LOGGER = getLogger(__name__)


def generate_convertion_data(
    board: Board,
    dac_steps: int,
    stats: list = [],
    minimum_r2: float = 0.9998,
    minimum_range: int = 10,
):
    """Add calibration data to the board

    Adds calibration data to the board for adc2mV conversion.

    Args:
        board: good old board
        dac_steps (int): amount of steps to divide the sweep in.
        stats: storage for progress messages
        minimum_r2:
        minimum_range:

    Raises:
        Unknown for now.
    """

    dsc = DACSweepController(board)
    adclr = ADCLinearReg(board)
    try:
        dsc.progress = stats
        adclr.progress = stats
    except:
        stats.append((0, "Calibration failed"))
        return
    dac_step = calc_step_size(board, dac_steps)
    LOGGER.debug("Will run calibration with step size: %s", dac_step)
    output2analyze = dsc.dac_sweep(step_size=dac_step)
    LOGGER.debug("Run linear regression on DAC sweep data: %s", len(output2analyze))
    caldata = adclr.linear_regression(output2analyze, minimum_r2, minimum_range)
    LOGGER.debug("Calibration data is: %s", caldata)
    board.caldata = caldata


def calc_step_size(board, steps):
    """Calculate the dac sweep step size based on amount of steps.

    Args:
        board: good ol' board object
        steps: amount of steps
    """

    max_counts = board.params.get("ext_dac", 0).get("max_counts", 0)
    min_counts = 0

    return (max_counts - min_counts) // steps



"""Preprocess raw data into events.

This module is used to convert raw data into a processed events.


"""
import logging
from copy import deepcopy

import numpy as np

from naludaq.helpers.exceptions import BadDataError
from naludaq.parsers import get_parser
from naludaq.tools.adc2mv.adc_converter import ADC2mVConverter
from naludaq.tools.pedestals.pedestals_correcter import PedestalsCorrecter
from naludaq.tools.timing_cal import TimingCorrecter

LOGGER = logging.getLogger(__name__)


class Preprocess:
    def __init__(self, board):
        """Proprocessor will convert raw events to preprocessed events

        Preprocessing includes parsing, pedestals correct, mV conversion, and timing calibration.

        Args:
            board: class object containing params, timingcal, pedestals, and/or adc2mvcal. This
                can also be an acquisition.
        """
        self.board = board

    @property
    def board(self):
        return self._board

    @board.setter
    def board(self, value):
        self._board = value
        self.params = getattr(value, "params", None)
        self.timingcal = getattr(value, "timingcal", None)
        self.pedestals = getattr(value, "pedestals", None)
        self.adc2mvcal = getattr(value, "caldata", None)
        self.samplingrate = self._get_samplerate()
        self.parser = get_parser(self.params)
        self.adc2mvconverter = ADC2mVConverter(self.params, self.adc2mvcal)
        self.pedestals_correcter = PedestalsCorrecter(self.params, self.pedestals)
        self.timing_correcter = TimingCorrecter(self.timingcal)

    def run(
        self,
        to_process: dict,
        correct_pedestals: bool = False,
        convert_adc2mv: bool = False,
        correct_timing: bool = False,
        convert_samples2time: bool = False,
        process_in_place: bool = False,
    ):
        """Process the raw event, parse and correct both x and y-axis

        Will parse raw data if no parsed exist.
        """
        if process_in_place is False:
            event = deepcopy(to_process)  # Sad copy to avoid overwriting event data
        else:
            event = to_process

        event = self._get_parsed_event(event)
        event = self._correct_yaxis(event, correct_pedestals, convert_adc2mv)
        event = self._correct_xaxis(event, correct_timing, convert_samples2time)

        return event

    def _get_parsed_event(self, to_process):
        """Parse event if it doesn't contain parsed data.

        Will check the event for parsed data and parse it if there is none.
        Event is a dict with a rawdata and data fields.

        Args:
            to_process: event to try and parse

        Returns:
            parsed event with a 'data' key.

        Raises:
            Doesn't raises errors but will return unparsed if parsing fails.
        """
        event = to_process
        if "data" not in event:
            try:
                event = self.parser.parse(to_process)
            except (Exception, BadDataError) as e_msg:
                LOGGER.error("process_event didn't work due to %s", e_msg)
        return event

    def _correct_yaxis(self, event, correct_pedestals: bool, correct_mv: bool):
        """Convert and correct the values on the Y-axis.

        Can either pedestals correct Y-axis or convert ADC counts to mV.
        Future upgrade would allow both.

        ## Important:
        Currently ADC to mV will run before pedestals correction. This means if the pedestals
        correction and convert_mV are enabled at the same time ONLY mV convertion will run!.

        Args:
            event(dict): Must contain a `data` field
            correct_mv(bool): True to enable ADC to mV conversion.
            correct_pedestals(bool): True to pedestals correct
        """
        if correct_mv:
            event = self.adc2mvconverter.run(event)
        elif correct_pedestals and self.pedestals is not None:
            event = self.pedestals_correcter.run(event)
        return event

    def _correct_xaxis(self, event, correct_time: bool, convert_time: bool):
        """Convert and timing correct values on time axis"""
        if correct_time:
            event["time"] = self.timing_correcter.run(event)

        if convert_time:
            # event = TimeConverter(self.board.params).convert(event)
            event["time"] = self._convert_sp_to_ns(event)
        return event

    def _convert_sp_to_ns(self, event):
        """Convert the time field in the event from samples to time.

        Using the sampling rate parameter

        Args:
            event

        Returns:
            Event with converted time if there is a time key, else just the event
        """
        xaxis = np.array(event.get("time", None))
        samplerate = self.samplingrate
        # np.arrays can't directly be checked against None, use `==`
        if np.any(xaxis == None) or samplerate is None:
            return xaxis

        return xaxis / samplerate

    def _get_samplerate(self):
        """Quick access function to sampling rate params"""
        return self.params.get("samplingrate", None)



"""Readout settings for the board.

Controls the readout channels and the readout windows.
"""
from logging import getLogger

from .default import ReadoutController

LOGGER = getLogger("naludaq.readout_controller_default")
MAX_READOUT = 65536


class TrbhmReadoutController(ReadoutController):
    """Special readout controller for the TRHBM.

    Overrides the channel-setting logic.
    """

    def __init__(self, board):
        super().__init__(board)

    def set_readout_channels(self, channels_to_read: list):
        """Select channels to readout.

        TRBHM currently doesn't support setting channels individually
        across both ASICs; their 'chansel' is shared.

        Args:
            channels_to_read(list): List of channel numbers to read
                Accepts channels 1..8 for consistent API, but does
                a logical OR between the upper and lower 4 channels.
        """
        self._validate_channels_or_raise(channels_to_read)

        chansel_lower = sum([1 << x for x in channels_to_read if x < 4])
        chansel_upper = sum([1 << (x - 4) for x in channels_to_read if x >= 4])
        self._write_control_register("chansel", chansel_lower | chansel_upper)

    def get_readout_channels(self) -> "list[int]":
        """Get the enabled readout channels as a list[int]."""
        chansel = self.board.registers["control_registers"]["chansel"]["value"]
        chansel = f"{chansel:02b}"
        return [c for c in range(self.board.channels) if chansel[c % 4] == "1"]



"""Light DAQ, lightweight data acquisition doing the bare minimum.
"""
import logging
import time
from collections import deque
from functools import partial
from queue import Queue
from threading import Lock, Thread

import numpy as np

from naludaq.backend.exceptions import BackendError
from naludaq.daq.workers import get_usb_reader
from naludaq.daq.workers.answer_parser_worker import AnswerParserWorker
from naludaq.daq.workers.csv_storage_worker import CSVStorageWorker
from naludaq.daq.workers.packager.worker_packager import OldPackagerLight, PackagerLight
from naludaq.daq.workers.packager.worker_packager_hdsoc import HDSoCPackager
from naludaq.models import Acquisition

LOGGER = logging.getLogger(__name__)


class LightDaq:
    """Data Acquisition (DAQ) for the ASoC.

    Capture and packetize bulk data from the Board.


    Args:
        board_controller: the board controller
        location: storage location for the data.


    Runs the different part of the acquisition pipeline.
    Serial reader
    Packager

    Attributes:
        parsed_output (bool): will output be parsed or raw. Default: True
        storage_directory (path): Where to store data to disk.
        frequency_packager (int): Polling speed of packager.
    """

    MAX_DISK_WRITERS = 1

    def __init__(self, board, location, start_workers):
        if board.using_new_backend:
            raise BackendError("DAQs are incompatible with the new backend")
        self._board = board

        self.storage_directory = location
        self._storage_directory = None
        self._output_buffer = deque()
        self.parsed_output = False
        self.answers = {}
        self.questions = {}
        self.answers_lock = Lock()
        self.serial_reader = None
        self.parser_worker = None
        self.disk_worker = None
        self.worker_packager = None
        self.answer_parser = None
        self._worker_serial_reader = None
        self._parsed_events = None  # Output
        self._acquisition_name = None
        self._frequency_packager = 50
        self.pkg_num = 0
        self.store_to_disk = False
        self.workers = list()

        # self.preprocess = None

        self._init_buffers()
        self._init_storage_settings()

        # if start_workers:
        #     self.init_and_start_workers()

    @property
    def storage_directory(self):
        """Get/Set the storage location for data saved to disk."""
        return self._storage_directory

    @storage_directory.setter
    def storage_directory(self, value):
        self._storage_directory = value

    @property
    def storage_settings(self):
        """Get/Set the storage location for data saved to disk."""
        return self._storage_settings

    @storage_settings.setter
    def storage_settings(self, value: dict):
        """Updates the storage settings used in preprocessing.
        Will only update values that are given.

        Args:
            value (dict):  {
            "correct_pedestals": bool
            "convert_mv": bool
            "correct_time": bool
            "convert_time": bool
            "process_in_place": bool

            }
        """
        self._storage_settings.update(value)
        if self.disk_worker:
            self.disk_worker.storage_settings = self.storage_settings

    @property
    def storage_namepattern(self):
        """Get/Set the storage namepattern.

        event_{} will fill the event number
        """
        return self._storage_namepattern

    @storage_namepattern.setter
    def storage_namepattern(self, value):
        self._storage_namepattern = value

    @property
    def storage_amount(self):
        """Get/Set the storage max amount."""
        return self._storage_amount

    @storage_amount.setter
    def storage_amount(self, value):
        self._storage_amount = value

    @property
    def acquisition_name(self):
        """Get/Set the Acquisition name."""
        return self._acquisition_name

    @acquisition_name.setter
    def acquisition_name(self, name):
        self._acquisition_name = name

    @property
    def output_buffer(self):
        """Get/Set the output buffer of the DAQ.

        This is where the daq store the output data.

        Args:
            output_buffer(deque, acquisition):
        """
        return self._output_buffer

    @output_buffer.setter
    def output_buffer(self, output):
        if not isinstance(output, (deque, Acquisition)):
            raise TypeError(
                "outputbuffer must be either a deque or an Acquisition obj."
            )
        self._output_buffer = output
        if self.disk_worker:
            self.disk_worker.input_buffer = self._output_buffer

    @property
    def board(self):
        """Change the board controller used.

        Useful if the board model is changed, then update the board controller.
        This can also be used to restart the daq since all Threads and Processes are stopped
        before changing the board controller.

        Args:
            board_controller: New board controller to use.
        """
        return self._board

    @board.setter
    def board(self, board):
        LOGGER.debug("DAQ: changing board to: %s", board.params["model"])
        self.stop_workers()
        self._board = board

        self.init_and_start_workers()

    def init_and_start_workers(self):
        """Starts the worker threads and initilizes queues and processes."""
        self._init_buffers()
        # self.setup_pipeline()

    def _init_buffers(self):
        """Initialize all inter-worker buffers used.

        There is a buffer between every worker.
        Buffers are made up of deque or queues.
        """
        # serial reader appends to:
        self._serial_to_packager_buffer = deque()
        self._packager_to_parser_buffer = deque()

        self._packager_to_answer_buffer = Queue()

    def _init_storage_settings(self):
        """Initializes the storage settings, used in Preprocessing."""
        self._storage_settings = {
            "correct_pedestals": False,
            "convert_adc2mv": False,
            "correct_timing": False,
            "convert_samples2time": False,
            "process_in_place": False,
        }

    def setup_pipeline(self):
        """Setup the components of the pipeline.

        Automagically starts the parts of the pipeline.
        """
        # self.init_packager()

        # self._init_parser()

        # Start Storagers (both)
        # self._init_store_raw_to_disk()
        # self._init_store_parsed_to_disk()

    def init_serial_reader(self):
        """Serial reader is started when data capture is started."""
        self._worker_serial_reader = self._create_serial_reader()
        self._worker_serial_reader.start()

    def _create_serial_reader(self):
        ser = get_usb_reader(
            self.board,
            self._serial_to_packager_buffer,
        )
        return ser

    def init_packager(self):
        """Initialize the basic workers.

        Serial reader and the packager are the first
        two steps in the data acquisition pipeline.
        """

        self.worker_packager = self._create_packager()
        self.worker_packager._pkg_num = self.pkg_num

        self.workers.append(self.worker_packager)
        self.worker_packager.start()

    def _create_packager(self):
        """Creates a packager object and connects it to the various I/O buffers.

        HDSoC gets a special packager.
        """
        kwargs = {
            "input_buffer": self._serial_to_packager_buffer,
            "output_buffer": self.output_buffer,
            "output_answers": self._packager_to_answer_buffer,
            "stop_word": self._board.params["stop_word"],
            "frequency": self._frequency_packager,
        }
        return {
            "hdsocv1": partial(HDSoCPackager, self._board),
            "hdsocv1_evalr1": partial(HDSoCPackager, self._board),
            "hdsocv1_evalr2": partial(HDSoCPackager, self._board),
        }.get(self.board.model, PackagerLight)(**kwargs)

    def init_diskstorage(self):
        """Create the disk storage.
        Only create it if it didn't already exist.
        """
        if not self.disk_worker:
            self.disk_worker = self._create_diskstorage()
        self.disk_worker.start()

    def _create_diskstorage(self):
        """ """
        return CSVStorageWorker(
            self.output_buffer,
            self.board,
            self.storage_settings,
            self.storage_directory,
        )

    def init_answer_parser(self):
        """Create the answer parser.

        Parsers the answers to regreads into a buffer.
        """
        self.answer_parser = self._create_answer_parser()

        self.workers.append(self.answer_parser)
        self.answer_parser.start()

    def _create_answer_parser(self):
        return AnswerParserWorker(
            self._packager_to_answer_buffer,
            self.answers,
            self.answers_lock,
        )

    def stop_workers(self):
        """Worker cleanup, shutdown the pipeline."""

        LOGGER.debug("Stopping threads")
        while self.workers:
            worker = self.workers.pop()
            LOGGER.debug("stopping: %s", worker)
            worker.stop()
            worker.join(5)

    def start_capture(self, acq_name="command"):
        """Start capturing serial data.

        Start polling the serial port also start storing to disk
        if the storeage options are set.
        """
        # Need to clear the worker buffers. Prevents bad problems!
        self._init_buffers()

        time.sleep(0.05)
        LOGGER.debug("STORE TO DISK: %s", self.store_to_disk)
        if self.store_to_disk:
            self.init_diskstorage()
        self.init_packager()
        self.init_answer_parser()
        self.init_serial_reader()

    def stop_capture(self):
        """Stops the capturing of data via the interface."""
        # self.stop_workers()
        if self._worker_serial_reader:
            self._worker_serial_reader.stop()
            self._worker_serial_reader.join()

        if self.worker_packager:
            self.pkg_num = self.worker_packager.stop()
            self.worker_packager.join()

        if self.answer_parser:
            self.answer_parser.stop()
            self.answer_parser.join()

        if self.disk_worker:
            self.disk_worker.stop()

        self.clear_workers()

        try:
            self._board.connection.reset_input_buffer()
        except Exception as error_msg:
            LOGGER.debug(
                "No connection to stop when stopping capture due to: %s", error_msg
            )

        # self._parsed_events = self.parse_all()

    def clear_workers(self):

        self._worker_serial_reader = None
        self.worker_packager = None
        self.answer_parser = None
        self.disk_worker = None

    def get_buffer_levels(self):
        """Return the amount in each of the buffers."""

        return {
            "serial buffer": len(self._serial_to_packager_buffer),
            "package buffer": len(self._packager_to_parser_buffer),
            "output": len(self.output_buffer),
        }

    def switch_output_buffer(self, new_buffer):
        """Switching the output buffer, the daq will finish the last batch
        before switching.

        Args:
            new_buffer: Acquisition or deque to append to.
            pedestals: Pedestals to use with next acquisition.
        """

        self.output_buffer = new_buffer

        if "name" in new_buffer.__dir__():
            self.acquisition_name = new_buffer.name

    def peek_latest_event(self) -> dict:
        """Returns the latest event parsed.

        Parses events on the fly, useful for oscilloscope plotting.

        Returns:
            Parsed event (dict) and pedestals corrected if requested.

        Raises:
            IndexError if the output_buffer is empty
        """
        event = None
        try:
            event = self.output_buffer[-1]
        except IndexError as emsg:
            raise IndexError(
                "Couldn't peek latest event %s, buffers: %s",
                emsg,
                self.get_buffer_levels(),
            )

        return event

    def read_register(self, reg_num: int, timeout: int = 10000):
        """Read a register by number.

        Ask the board for a register and wait until there is a response.

        Args:
            reg_num(int): The register to read. See board specific register map.

        Returns:
            The register
        """
        # Make sure the id is unique.
        while True:
            cmd_id = np.random.randint(0, 2**16)
            if self.questions.get(cmd_id, False) is False:
                self.questions[cmd_id] = ""
                LOGGER.debug("Asked question: %s", cmd_id)
                break

        command = "AD{0:02x}{1:04x}".format(reg_num, cmd_id)

        self.board.connection.send(command)

        def find_answer(cmd_id, answer_lock, answers, answer):
            timeout = 5
            t_count = 0
            while True:
                print("waiting for answer")
                with answer_lock:
                    print("inside the lock")
                    answer = answers.get(cmd_id, None)
                    print(f"Available answers: {answers}")
                if answer:
                    break
                time.sleep(0.5)
                t_count += 1
                if t_count >= timeout:
                    break
            print(f"--------got an answer: {answer}")

        answer = ""
        finder = Thread(
            target=find_answer, args=(cmd_id, self.answers_lock, self.answers, answer)
        )
        LOGGER.debug("waiting to find answer")
        finder.daemon = True
        finder.start()
        finder.join()
        LOGGER.debug("Answer found: %s", answer)
        return answer


#####################################################################


class OldLightDaq:
    """Data Acquisition (DAQ) for the ASoC.

    Args:
        board_controller: the board controller
        location: storage location for the data.


    Runs the different part of the acquisition pipeline.
    Serial reader
    Packager

    Attributes:
        parsed_output (bool): will output be parsed or raw. Default: True
        storage_directory (path): Where to store data to disk.
        frequency_packager (int): Polling speed of packager.
    """

    MAX_DISK_WRITERS = 1

    def __init__(self, board):
        self._board = board
        self.serial_reader = None
        self.worker_packager = None
        self._worker_serial_reader = None
        self._parsed_events = None  # Output
        self._acquisition_name = None
        self._frequency_packager = 1000
        self._pkg_num = 0

        self.workers = []

        self._init_buffers()

    @property
    def storage_directory(self):
        """Get/Set the storage location for data saved to disk."""
        return self._storage_directory

    @storage_directory.setter
    def storage_directory(self, value):
        self._storage_directory = value

    @property
    def acquisition_name(self):
        """Get/Set the Acquisition name."""
        return self._acquisition_name

    @acquisition_name.setter
    def acquisition_name(self, name):
        self._acquisition_name = name

    @property
    def output_buffer(self):
        """Get/Set the output buffer of the DAQ.

        This is where the daq store the output data.

        Args:
            output_buffer(deque, acquisition):
        """
        return self._output_buffer

    @output_buffer.setter
    def output_buffer(self, output):
        if not isinstance(output, (deque, Acquisition)):
            raise TypeError(
                "outputbuffer must be either a deque or an Acquisition obj."
            )
        self._output_buffer = output

    @property
    def board(self):
        """Change the board controller used.

        Useful if the board model is changed, then update the board controller.
        This can also be used to restart the daq since all Threads and Processes are stopped
        before changing the board controller.

        Args:
            board_controller: New board controller to use.
        """
        return self._board

    @board.setter
    def board(self, board):
        LOGGER.debug("DAQ: changing board to: %s", board.params["model"])
        self.stop_workers()
        self._board = board

        self.init_and_start_workers()

    def init_and_start_workers(self):
        """Starts the worker threads and initilizes queues and processes."""
        self._init_buffers()
        # self.setup_pipeline()

    def _init_buffers(self):
        """Initialize all buffers used.

        There is a buffer between every worker.
        Buffers are made up of deque or queues.
        """
        # serial reader appends to:
        self._serial_to_packager_buffer = deque()
        self._packager_to_parser_buffer = deque()

        # Output buffer
        self.output_buffer = deque()

    def setup_pipeline(self):
        """Setup the components of the pipeline.

        Automagically starts the parts of the pipeline.
        """
        # self.init_packager()

        # self._init_parser()

        # Start Storagers (both)
        # self._init_store_raw_to_disk()
        # self._init_store_parsed_to_disk()

    def init_serial_reader(self):
        """Serial reader is started when data capture is started."""
        self._worker_serial_reader = self._create_serial_reader()
        self._worker_serial_reader.start()

    def _create_serial_reader(self):
        return get_serial_reader(self.board, self._serial_to_packager_buffer)

    def init_packager(self):
        """Initialize the basic workers.

        Serial reader and the packager are the first
        two steps in the data acquisition pipeline.
        """

        self.worker_packager = self._create_packager()

        self.workers.append(self.worker_packager)
        self.worker_packager.start()

    def _create_packager(self):

        return OldPackagerLight(
            self._serial_to_packager_buffer,
            self.output_buffer,
            self._board.params["stop_word"],
            self._frequency_packager,
        )

    def stop_workers(self):
        """Worker cleanup, shutdown the pipeline."""

        LOGGER.debug("Stopping threads")
        while self.workers:
            worker = self.workers.pop()
            LOGGER.debug("stopping: %s", worker)
            worker.stop()
            worker.join(5)

    def start_capture(self, acq_name="command"):
        """Start capturing serial data.

        Start polling the serial port also start storing to disk
        if the storeage options are set.
        """
        time.sleep(0.05)
        self.init_packager()
        self.init_serial_reader()

    def stop_capture(self):
        """Stops the capturing of data via the interface."""
        # self.stop_workers()
        if self._worker_serial_reader:
            self._worker_serial_reader.stop()
            self._worker_serial_reader.join()

        if self.worker_packager:
            self.worker_packager.stop()
            self.worker_packager.join()

        self.clear_workers()

        try:
            self._board.connection.reset_input_buffer()
        except Exception as error_msg:
            LOGGER.debug(
                "No connection to stop when stopping capture due to: %s", error_msg
            )

        # self._parsed_events = self.parse_all()

    def clear_workers(self):

        self._worker_serial_reader = None
        self.worker_packager = None

    def get_buffer_levels(self):
        """Return the amount in each of the buffers."""

        return {
            "serial buffer": len(self._serial_to_packager_buffer),
            "package buffer": len(self._packager_to_parser_buffer),
            "output": len(self.output_buffer),
        }

    def switch_output_buffer(self, new_buffer):
        """Switching the output buffer, the daq will finish the last batch
        before switching.

        Args:
            new_buffer: Acquisition or deque to append to.
            pedestals: Pedestals to use with next acquisition.
        """

        self.output_buffer = new_buffer

        if "name" in new_buffer.__dir__():
            self.acquisition_name = new_buffer.name



import logging
from collections import deque

from naludaq.helpers.exceptions import OperationCanceledError

logger = logging.getLogger("naludaq.operations")


class ProgressReporter:
    def __init__(self, *args, **kwargs):
        """Base class for operations that report progress."""
        super().__init__(*args, **kwargs)
        self._progress = []

    @property
    def progress(self):
        """Get/set the object that stores the progress updates.

        The object must have an 'append' method.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError(
                "Progress updates are stored in an object with an 'append' method"
            )
        self._progress = value

    def update_progress(self, percent: float, message: str):
        """Updates the progress with a percent and message.

        Args:
            percent (float): the percent completion of the task
            message (str): a description of what is currently taking place
        """
        logger.debug("%s: %s | %s", type(self).__name__, percent, message)
        if isinstance(self.progress, (list, deque)):
            self.progress.append((percent, message))
        else:
            try:
                self.progress.update_status(percent, message)
            except:
                raise TypeError(
                    "Progress updates are stored in an object with an 'append' method"
                )


class CancelableOperation:
    def __init__(self, *args, **kwargs):
        """Base class for operations that can be canceled."""
        super().__init__(*args, **kwargs)
        self._canceled = False

    @property
    def canceled(self):
        """Get whether the operation has been canceled."""
        return self._canceled

    def cancel(self):
        """Cancel the operation."""
        self._canceled = True

    def _raise_if_cancelled(self):
        """Raises an OperationCanceledError if the operation has been canceled."""
        if self.canceled:
            raise OperationCanceledError()



"""
Driver for talking over a FT60X interface interface

Should be interoperable with the linketh package


Marcus Luck - Nalu Scientific
email: marcus@naluscientific.com
January 2023
"""
import binascii
import ctypes
import logging
import platform
import time
from typing import Optional

from naludaq.helpers import validations

try:
    import ftd3xx as ftd
except (FileNotFoundError, OSError):
    raise FileNotFoundError("D3XX driver not installed")
except ImportError:
    raise ImportError("ftd3xx not installed. Please install package.")

if platform.system() == "Windows":
    import ftd3xx._ftd3xx_win32 as _ft
# elif platform.system() == "Linux":
#     import ftd3xx._ftd3xx_linux as _ft
else:
    raise ImportError(f"Support for USB 3 on {platform.system()} is unavailable")

import numpy as np

from naludaq.board.connections.base_serial import BaseConnection
from naludaq.helpers.exceptions import FTDIError
from naludaq.parsers.answer_parser import get_answer_parser

logger = logging.getLogger("naludaq.D3XX")
_WRITE_MESSAGE_NIBBLES = 8
_DEFAULT_READ_PIPE_ID = 0x82
_DEFAULT_WRITE_PIPE_ID = 0x02
_DEFAULT_CHUNK_SIZE = 4 * 1024  # 786_842


class USB3(BaseConnection):
    """FTDI is a communications class for talking to the boards.

    Uses the ftd2xx Python package, which requires d2xx drivers
    to be properly installed. For information about using d2xx:
    https://ftdichip.com/wp-content/uploads/2020/08/D2XX_Programmers_GuideFT_000071.pdf
    """

    def __init__(self, conn_info: dict):
        """Initializes the FTDI object, but does not open a connection.

        Args:
            conn_info (dict): the connection info dict. Required keys: 'speed', 'model',
                and 'serial_number'. 'stop_word' is preferred, and defaults
                to 0xCAFE if not present.

        Raises:
            ConnectionError if the connection info dict is missing the required keys.
        """
        super().__init__()
        self._conn: ftd.FTD3XX = None
        self._read_addr = "AD"
        self._write_addr = "AF"
        self._is_open = False

        self._chunk_size = _DEFAULT_CHUNK_SIZE
        self._read_pipe_id = _DEFAULT_READ_PIPE_ID
        self._write_pipe_id = _DEFAULT_WRITE_PIPE_ID
        self.tx_pause = 0.02
        self.bundle_mode = False

        self._speed = conn_info["speed"]
        self._model = conn_info["model"]
        self.stopword = conn_info.get("stop_word", b"\xca\xfe")  # b'\xfa\xce'
        self.new_firmware = conn_info.get("new_firmware", False)
        self.response_length = 4
        self.parse_answer = get_answer_parser()

        serial_number = conn_info.get("serial_number", None)
        if serial_number:
            logger.debug("using serial_number %s to connect", serial_number)
            self.serial_number = serial_number
        else:
            raise ConnectionError("FTDI connection requires the serial num or port num")

    def __del__(self):
        """Closes the connection if one is open."""
        self.close()

    @property
    def new_firmware(self):
        """Get/Set if the board uses a new firmware or not.

        If a new firmware is used, the register answers are different.
        """
        return self._new_firmware

    @property
    def in_waiting(self):
        return self._chunk_size

    @new_firmware.setter
    def new_firmware(self, value: bool):
        self._new_firmware = value
        self.response_length = 4
        if value is True:
            self.response_length = 8
        self.parse_answer = get_answer_parser(value)

    @property
    def is_open(self) -> bool:
        """Indicates whether the connection is open."""
        is_open = True
        try:
            self._conn.getChipConfiguration()
        except (AttributeError, ftd.DeviceError):
            is_open = False
        return is_open

    def open(self):
        """Open connection over the FTDI interface.

        Set board (device) number and baud before opening a connection, use:
            .serial_number = value
        """
        if self.is_open:
            logger.debug("USB3.open(): device already open")
            return

        serial_number = self.serial_number.encode("ascii")
        usb_connection = ftd.create(serial_number, _ft.FT_OPEN_BY_SERIAL_NUMBER)
        self._conn = usb_connection
        self._is_open = usb_connection is not None

        if usb_connection is None:
            raise ConnectionError("Failed to connect USB3")

    def close(self):
        """Close the serial port."""
        if self._conn is not None:
            try:
                self._conn.close()
            except Exception as e:
                logger.debug("Failed to close connection due to: %s", e)
            self._conn = None
        self._is_open = False

    def send(self, data, pause: Optional[float] = None, bundle: Optional[bool] = None):
        """Converts the string to a series of bytes to send in the correct format

        Input:
            data (hex): A hex type string (interface needs 32 bits, so 8 hex chars minimum)
            pause (float): How logn to wait in between send.
            bundle (bool): If true, will send longer packages in one send, otherwise will send cmd by cmd.
        """
        if bundle is None:
            bundle = self.bundle_mode
        if pause is None:
            pause = self.tx_pause

        if len(data) % _WRITE_MESSAGE_NIBBLES != 0:
            logger.debug(
                "FTDI.send(): need multiples of 8 hex chars, you gave me %s", len(data)
            )
            logger.debug("command:%s", data)
            return

        logger.debug("Sending: %s in %s mode", data, "bundle" if bundle else "single")
        if (len(data) > _WRITE_MESSAGE_NIBBLES) and bundle is False:
            for i in range(0, len(data), _WRITE_MESSAGE_NIBBLES):
                self._send(data[i : i + _WRITE_MESSAGE_NIBBLES], pause)
        else:
            self._send(data, pause)

    def _send(self, data, pause):
        try:
            tosend = binascii.a2b_hex(data)
        except binascii.Error as e:
            logger.error(f"FTDI.send(): could not convert hex data: {e}")
            return
        try:
            bytes_written = self._conn.writePipe(
                self._write_pipe_id, tosend, len(tosend)
            )
        except AttributeError as error_msg:
            logger.error(
                "Can't send command to board, no connection available. %s", error_msg
            )
        if bytes_written == 0:
            raise ConnectionError("Failed to send data, no connection available")
        time.sleep(pause)

    def read_until(self, stopword: bytes) -> bytearray:
        """Pulls data from the connection until a set of characters is found.

        Args:
            stopword (bytes): the sequence of characters that indicate
                when to stop reading.

        Returns:
            A bytearray containing the data, including the trailing stop word.
            The buffer is empty or incomplete if timed out
        """
        lenstop = len(stopword)
        buff = bytearray()
        while True:
            value = None
            try:
                value = self.read(self._chunk_size)
            except Exception as e:
                logger.debug("Receiving data failed due to: %s", e)
            if value is None or len(value) == 0:
                logger.error("Failed to read data")
                break

            buff += value
            if buff[-lenstop:] == stopword:
                break

        logger.debug("Received: %s", buff.hex())

        return buff

    def receive(self):
        """Reads out an event or answer. Both end with a stopword.

        This will *NOT* work on boards where data and registers have different stopwords.
        """
        return self.read_until(self.stopword)

    def read(self, length: int = None) -> bytes:
        """Attempt to read data from the connection.

        Args:
            length (int, optional): maximum number of bytes to read.
                If not provided, the chunk size is assumed.

        Raises:
            FTDIError: If the operation failed.

        Returns:
            bytes: The data received. This may be less than the amount requested.
        """
        if length is None:
            length = self._chunk_size
        try:
            data = ctypes.c_buffer(length)
            rxlength = self._conn.readPipe(self._read_pipe_id, data, length)
            data = data[0:rxlength]
        except ftd.DeviceError as e:
            raise FTDIError(f"USB.read failed due to: {e}")
        if data == b"\xFA\xCE":
            return b""
        return data

    def read_all(self, attempts=10) -> bytearray:
        """Read all waiting bytes"""
        data = bytearray()

        attempt_count = 0
        while attempt_count < attempts:
            resp = self.read(self._chunk_size)
            if len(resp) == 0:
                attempt_count += 1
                continue
            data += resp
        return data

    def reset_input_buffer(self, iterations: int = 100):
        """Discards any data currently in the boards output buffer.

        This will empty both the computer input buffer and the hardware
        output buffer.

        However, this will only run 100 times at the most which is ~400kb.

        Args:
            iterations (int): number of times to read the buffer.
                A cap is necessary since the board may be sending data faster
                than it can be purged.
        """
        validations.validate_positive_int_or_raise(iterations)
        for _ in range(iterations):
            if len(self.read_all()) == 0:
                break
            time.sleep(0.001)

    def reset_output_buffer(self):
        """Not implemented for USB3."""

    def set_answer_parser(self, old=True):
        """Set's the answer parser"""
        self.parse_answer = get_answer_parser(old)

    def readReg(self, regNum, timeout=10000) -> int:
        """
        Reads out the register specified by the integer regNum
        Args:
            regNum (int): The registry number to read
            timeout (int): Amount of retries.
        Returns:
            Hex value of the serialread of two bytes.
        """
        # Send command
        self._send_readreg_cmd(regNum)

        # Wait for response
        buffer = self.receive()
        # Parse response
        answer = self.parse_answer(buffer)
        # Make sure responses are valid

        try:
            logger.debug(
                "FTDI.readReg(%s) success: %s %s %s",
                answer["read_reg"],
                answer["value"],
                hex(answer["value"]),
                "{0:b}".format(answer["value"]).zfill(16),
            )
        except Exception:
            pass

        return answer["value"]

    def _send_readreg_cmd(self, regNum):
        """Generates and sends a command to read a register.

        Args:
            regNum (str, int): address as int or hex str.
        """
        cmd_id = np.random.randint(0, 2**16)
        if isinstance(regNum, str):
            regNum = int(regNum, 16)
        read_addr = int(self._read_addr, 16)

        try:
            command = f"{read_addr:x}{regNum:02x}{cmd_id:04x}"
        except Exception as error_msg:
            logger.exception(f"send_readreg_cmd create cmd failed: {error_msg}")
        else:
            self.send(command)



"""Parser

Use parser to parse the data from the Hardware.
The boards sends a raw bitstream and it needs to be parsed into events.
The parser class is a tool to parse the raw bitstream of 8-bit chunks
into 16 bit words then extracting the 12-bit data, the header info,
channels and window number data.

This data is then returned in a dictionary.
The raw data is preserved.

This module was previously located under the naludaq.daq.workers
subpackage, and was relocated to improve the organization of the
naludaq package.
"""

import logging

import numpy as np

from naludaq.helpers.exceptions import BadDataError
from naludaq.parsers.headers import get_header_parser
from naludaq.parsers.parser import Parser

LOGGER = logging.getLogger(__name__)


class Aardvarcv3Parser(Parser):
    """Parser for the AARDVARCv3 data protocol.

    The AARDVARCv3 adds a windows header compared to the other protocol which
    causes bitlength to be an unreliable way to determine channels/windows in data.
    This parser implements ways to work around this in a consistent manner.

    ```
    Dataformat:
        - 3 x event headers
        - N x window packet
            - 1-4 x channels
                - 1 x channel hader
                - 64 x data packets
            - 1 x window footer
        - 1 x event footer
    ```
    Since the event package can contain 1-4 channels of data the window header position
    is hard to determine.

    The solution is to pre-generate all package length combinations and use as a lookup table.
    This allow us to quickly validate packages where channel amount is not ambiguous.

    IN the rare cases it's abiguous there is a test where it checks if window numbers are the same for
    each channel in a window package.

    """

    def __init__(self, params):
        self.params = params

        self._parse_event_headers = get_header_parser(self.params["model"])

        self.test_amount_windows = 3
        self.samples = self.params.get("samples", 64)
        self.windows = self.params.get("windows", 512)
        self.channels = self.params.get("channels", 4)
        self.num_evt_headers = self.params.get("num_evt_headers", 3)
        self.num_channel_headers = self.params.get("num_channel_headers", 1)
        self.num_window_headers = self.params.get("num_window_headers", 1)
        self.data_bitmask = self.params.get("data_bitmask", 8191)
        self.chanmask = self.params.get("chanmask", 3072)
        self.windmask = self.params.get("windmask", 1022)
        self.chanshift = self.params.get("chanshift", 10)
        self.num_footer_words = self.params.get("num_footer_words", 2)
        self.last_bits = self.params.get("lastbits", 1)
        self.chan_step_size = self.samples + self.num_channel_headers

        # Generatethe valid bits dictionary
        self.valid_bitlengths = self.create_valid_bitlengths()

    def _validate_input_package(self, in_data):
        """Validate the input data.

        Returns:
            chan (int): Number of channels in data
            wind (int): Number of windows in data

        Raises:
            BadDataException if:
            - there is no data.
            - there is not a en event amount of bytes, since 8-bits are combined to 16-bit words.
            - the data is not divisable by sample_num + window_headers.
            - the data is not a multiple of a combination of windows and channels
        """
        if not isinstance(in_data, dict):
            raise TypeError("raw_data is not a dict it's a %s", type(in_data))

        if in_data["rawdata"] == b"":
            raise BadDataError("No Data Found")

        # Edge case, data doesn't contain valid words.
        if len(in_data["rawdata"]) % 2 != 0:
            raise BadDataError(
                f"Bad package no {in_data.get('pkg_num', 'unknown')}, "
                f"length is not an even amount of words, "
                f"len {len(in_data['rawdata'])}"
            )
        return self.get_chan_wind_from_bitlength(in_data)

    def parse_digital_data(self, in_data):
        """Parse the raw data from the board.
        Assuming that the data packets are constant length, we can extract the data
        in place with matrix operations, speeding up the parsing.

        Parses the data spit out by the digitial readout portion of the firmware

        Channel data is interleaved per packet, e.g
        pkt0: [header, ch0, ch1, ch2, ch3, dig. ftr.]
        pkt1: [header, ch0, ch1, ch2, ch3, dig. ftr.]
        etc

        TODO: To speed this up more, we do not want two matrices (bitshift & raw),
        we would perferably only want one.

        Args:
            in_data (bytearray): Raw data from the board
        Returns:
            Parsed event as a dict.
        Raises:
            BadDataError if no data is found or if the data contains errors.

        Args:
            data (np.array or list): all the data, not multi dimensional so one channel at the time.

        """
        try:
            # make sure package contains data and there's an even amount of bytes to combine.
            chans, wind = self._validate_input_package(in_data)
        except:
            raise

        # TODO(Marcus): Need logic for the different channels. Since this needs to be processed from readout data.
        channels = self.channels
        num_headers = self.num_evt_headers
        windmask = self.windmask
        last_bits = self.last_bits
        chan_step_size = self.chan_step_size
        # Converts two bytes into 16-bit words, effectively swapping byteorder and combining in one.
        raw_data = np.frombuffer(in_data["rawdata"], dtype=">H")

        bitshifted_data = self._bitshift_data(raw_data)

        curr_event = {
            "window_labels": [[] for _ in range(channels)],
            "data": [[] for _ in range(channels)],
            "start_window": (raw_data[num_headers] & windmask) >> last_bits,
        }

        self._parse_event_headers(curr_event, bitshifted_data)

        # Packet size
        packet_size = self.chan_step_size * chans + self.num_window_headers

        # Num packets
        num_packet = wind

        data_end = len(bitshifted_data) - self.num_footer_words

        event_matrix = np.reshape(
            bitshifted_data[num_headers : data_end + 1],
            [packet_size, num_packet],
            order="F",
        )
        raw_event_matrix = np.reshape(
            raw_data[num_headers : data_end + 1], [packet_size, num_packet], order="F"
        )

        curr_event["window_timings"] = event_matrix[-1, :]
        for i in range(chans):
            curr_chan = (
                raw_event_matrix[i * chan_step_size, 0] & self.params["chanmask"]
            ) >> self.params["chanshift"]
            curr_event["data"][curr_chan] = (
                event_matrix[1 + i * chan_step_size : (i + 1) * chan_step_size, :]
            ).flatten(order="F")
            curr_event["window_labels"][curr_chan] = (
                raw_event_matrix[i * chan_step_size, :] & self.params["windmask"]
            ) >> self.last_bits

        return curr_event

    def get_chan_wind_from_bitlength(self, in_data):
        """Get channels and window amoutns from package length."""
        raw_data = np.frombuffer(in_data["rawdata"], dtype=">H")
        raw_data_length = len(raw_data)

        chanwinds = self.valid_bitlengths.get((raw_data_length), [])
        chan = wind = 0

        if not chanwinds:
            raise BadDataError(
                f"Bad package no {in_data.get('pkg_num', 'unknown')}, "
                f"the amount of bytes is not a valid length."
            )

        if len(chanwinds) == 1:
            chan = chanwinds[0][0]
            wind = chanwinds[0][1]

        elif len(chanwinds) > 1:  # potentials.get(rawlen-4, False):
            for chanwind in sorted(chanwinds, key=lambda x: x[0], reverse=True):
                chan = chanwind[0]
                wind = chanwind[1]

                if self.test_chanwinds(raw_data, chan):
                    break

        return chan, wind

    def test_chanwinds(self, raw_data, chans):
        """Check data integrity using the channels
        All windows should be the same in one block of channels.
        If they are not something is wrong, and the check will stop and return False.

        Args:
            raw_data is the the raw data byteshifted and bytes combined to 16-bit words.
            chans is the amount of channel to test for.
        Returns:
            True if test is passed, False on fail.

        """
        # PARAMS #################################

        step_size = self.chan_step_size * chans + self.num_window_headers
        num_evt_headers = self.num_evt_headers
        last_bits = self.last_bits
        pass_fail = False

        # runs through a predefined number of windows.
        header_steps = range(
            num_evt_headers, step_size * self.test_amount_windows, step_size
        )
        pass_fail_windows = [
            True
        ]  # Converting to set will fail unless the intial value is True.
        for idx in header_steps:
            # step size depends on channel

            windows = []
            window_steps = range(idx, idx + step_size - 1, self.chan_step_size)
            for inner_idx in window_steps:
                pkt = raw_data[inner_idx]
                window = pkt & self.windmask
                window = window >> last_bits
                windows.append(window)
                # check all windownums are the same
            if len(set(windows)) == 1:
                pass_fail_windows.append(True)
            else:
                pass_fail_windows.append(False)
                break

        if len(set(pass_fail_windows)) == 1:
            pass_fail = True

        return pass_fail

    def create_valid_bitlengths(self):
        """AARDVARCv3 valid bitlentgth combinations

        There are only certain amount of bytes that are valid for the aardvarcv3 protocol.

        Returns:
            Dictionary of valid combinations, {bytes: [(channels, windows), (channels, windows)]}
        """
        bitlengths = {}
        headers = 4
        for win in range(1, self.windows):  # winn never read out 0 windows
            for chan in range(1, self.channels + 1):
                packlen = (self.chan_step_size * chan + 1) * win + headers
                cur_value = bitlengths.get(packlen, [])
                cur_value.append((chan, win))
                bitlengths[packlen] = cur_value
        return bitlengths

    def get_conflicts(self):
        """Get all bitlengths with ambiguity.

        Args:
            bitlengths (dict):

        """
        potentials = {key: x for key, x in self.valid_bitlengths.items() if len(x) > 1}
        print(potentials)
        return potentials



from .base import HeaderParser


class UPAC32HeaderParser(HeaderParser):
    @staticmethod
    def parse_header(event: dict, raw_data: bytes):
        """Parse event headers from raw data and store them into an event dict.

        Args:
            event (dict): the event
            raw_data (bytes): the raw data to parse.
        """
        # event['prev_final_window'] = (raw_data[0]&255) # evt header 0
        # trigger_time_ns = raw_data[0] << 12 + raw_data[1] # 1
        # trigger_time_ns += raw_data[1] # 2 # 24 bit number
        event["trigger_time_ns"] = (raw_data[0] << 12) + raw_data[1]
        event["event_id"] = raw_data[2]  # 3



"""Python client for the Rust backend (naludaq_rs).

This package contains all the basic functionality required to communicate with the Rust backend.
Higher-level functionality (such as the DataCollector) is provided by the rest of the NaluDAQ package.

The Rust backend is an application which runs either locally (in-process or out-of-process) or remotely (over the network).
Its purpose is to provide a flexible and performant interface for controlling hardware and acquiring data.
It provides a REST API through HTTP for communicating with the hardware, which the `backend` subpackage interally uses
for everything. The user should never need to care about the specifics of the REST API.

There really very few things this client cares about:
- Hardware connections and I/O (abstracted as "Devices")
- Data (acquisition) storage and retrieval
- Server configuration

The connection to the server is held by a "Context," which provides a simple HTTP interface. The context
does not usually need to be created manually, as it is created by the `Board` object. However, if it is
desired to communicate with the server without a Board object then one can be created manually.

Control of the server is split into "managers", which are each responsible for managing a specific aspect
of the server. The only managers are:

- AcquisitionManager - data acquisition and retrieval
- BoardIoManager - reading and writing registers
- ConfigManager - server configuration (you probably won't need to use this often)
- ConnectionManager - opening and configuring connections to hardware

Note that the BoardIoManager, ConfigManager, and ConnectionManager will rarely need to be used
externally, as the Board object and various controllers provide a higher-level interface; the
AcquisitionManager is the only one the user should really care about.

Besides controlling the server, there are a few useful types which model the hardware and data:

- RemoteAcquisition - accessing an acquisition remotely
- LocalAcquisition - accessing an acquisition in memory
- DiskAcquisition - accessing an acquisition on disk
- TemporaryAcquisition - a remote acquisition, but only exists for as long as you need it
- Device - a representation of a hardware device, mainly useful for viewing available devices.
"""


from .context import Context
from .managers import (
    AcquisitionManager,
    BoardIoManager,
    ConfigManager,
    ConnectionManager,
)
from .models import (
    AvailableD2xxDevice,
    AvailableSerialDevice,
    D2xxDevice,
    Device,
    DeviceType,
    DiskAcquisition,
    DiskChunk,
    LocalAcquisition,
    RemoteAcquisition,
    SerialDevice,
    TemporaryAcquisition,
    UdpDevice,
)



import ipaddress

from naludaq.backend.managers.base import Manager


class ConfigManager(Manager):
    def __init__(self, board):
        """Utility for higher-level configuration of the backend.

        Args:
            context (Context): context used to communicate with the backend.
        """
        super().__init__(board)

    @property
    def is_embedded_server(self) -> bool:
        """Check if the server is running in the same process via the Python bindings."""
        return getattr(self.board, "server", None) is not None

    @property
    def is_server_local(self) -> bool:
        """Check if the server is running on the same machine as the current process.
        This is done by checking if the server is connected to over the loopback interface.
        """
        local = self.is_embedded_server
        if not local:
            try:
                local = ipaddress.ip_address(self.context.client.host).is_loopback
            except (TypeError, ValueError):
                # don't do anything, this is the wrong place to raise an error
                pass
        return local

    def configure_packaging(self, events: "bytes | str", answers: "bytes | str"):
        """Sets the stop words used by the backend to separate
        events and answers.

        Args:
            events (bytes | str): stop word for events.
            answers (bytes | str): stop word for answers (register reads).
        """
        if isinstance(events, bytes):
            events = events.hex()
        if isinstance(answers, bytes):
            answers = answers.hex()
        try:
            self.context.client.put(
                "/server/data-format",
                params={
                    "model": self.board.model,
                    "events": events,
                    "answers": answers,
                },
            )
        except ValueError as e:
            raise ValueError("Stop words are invalid") from e



from .acquisition import (
    DiskAcquisition,
    DiskChunk,
    LocalAcquisition,
    RemoteAcquisition,
    TemporaryAcquisition,
)
from .device import (
    AvailableD2xxDevice,
    AvailableD3xxDevice,
    AvailableSerialDevice,
    D2xxDevice,
    D3xxDevice,
    Device,
    DeviceType,
    SerialDevice,
    UdpDevice,
)



"""SerialRead is the first piece in a multithread double buffer.

OVERVIEW:
By reading the OS serial buffer with a certain polling speed and storing it in
an internal queue buffer overflows can be avoided if readout is fast.

This serial reader extends the QThread from pyside2.
Start the serial reader together with the package creator since the
serial buffer is raw bytes and not useful for the application before
the data has been packetized.

    Functions:
         run: starts the thread
         stop: stop the thread.

"""
import logging
from collections import deque
from threading import Thread

import serial

# Create logger for application
LOGGER = logging.getLogger(__name__)


class SerialReader(Thread):
    """Thread worker to read external serial buffer.

    Moves data from the OS serial buffer to the applications
    internal serial buffer. The internal buffer is a deque.

    Args:
        ser_con (serial): Serial connection.
        serial_buffer (deque): Threadsafe external buffer to write to.
        frequency (float): Deprecated, no need to supply anymore.

    Functions:
         start: starts the thread
         stop: stop the thread.

    """

    def __init__(self, connection, serial_buffer, frequency=None):
        super().__init__()
        self._interval = 0
        self._running = True
        self.daemon = True
        self._ctrl = connection
        self._serial_buffer = serial_buffer

    @property
    def serial_buffer(self):
        """Set the serial buffer.

        Serial buffer must be deque.

        Raises:
            TypeError if input is not an deque.
        """
        return self._serial_buffer

    @serial_buffer.setter
    def serial_buffer(self, value):
        if not isinstance(value, deque):
            raise TypeError("The serial_buffer must be a deque.")

        self._serial_buffer = value

    def read_serial_worker(self):
        """Worker takes raw serial data from driver buffer and stores in internal buffer.

        This solves issues with buffer overruns on higher UART baud rates.
        Runs with the polling frequency set, if to high it draws a lot of CPU power.
        If too low you get buffer overflow.

        Returns:
            appends finished packages to the package_buffer.

        """
        # Before starting to read an event, flush the buffer from whatever.
        self._reset_input_buffer()
        try:
            while self._running and self._ctrl.is_open:
                data = self._read_serial_data()
                self._store_serial_data(data)
        except KeyboardInterrupt:
            pass

    def _reset_input_buffer(self):
        self._ctrl.reset_input_buffer()

    def _read_serial_data(self):
        data = None
        try:
            # read all that is there or wait for one byte (blocking)
            data = self._ctrl.read(self._ctrl.in_waiting or 1)
        except (serial.SerialException, AttributeError) as error_msg:
            # probably some I/O problem such as disconnected USB serial
            LOGGER.error(error_msg)
            self._running = False
            data = None
        return data

    def _store_serial_data(self, data):
        if data:
            # make a separated try-except for called user code
            try:
                self._serial_buffer.appendleft(data)
            except Exception as error_msg:
                LOGGER.error(error_msg)
                self._running = False

    def run(self):
        """Start the thread worker.

        Runs the read_data continously.
        Gathers events from serial port and store in buffer.
        """
        self._running = True
        self.read_serial_worker()

    def stop(self):
        """Stops the thread worker.

        Sets running to False and stops the loop.
        """
        LOGGER.debug("Serial worker thread stopping, stop command received")
        self._running = False



import time
from typing import Dict

import numpy as np

from naludaq.communication import ControlRegisters
from naludaq.controllers.trigger import get_trigger_controller
from naludaq.tools.threshold_scan.threshold_scan import ThresholdScan


class Upac96ThresholdScan(ThresholdScan):
    """Scan each trigger value and read the scaler values for each channel."""

    def _get_scalar_values(self, pause: float) -> np.ndarray:
        """Scan the range and return np.array with trigger amounts.

        Args:
            pause (float): amount of seconds to pause in between samples.

        Returns:
            Triggers value per channel as `np.array`.
        """
        trigger_select_backup = self._get_ctrl_reg("trigger_select")
        self._set_ctrl_reg("trigger_select", 1)
        scan_values = self.scan_values
        output = np.zeros((self.board.channels, len(scan_values)))

        if self.board.using_new_backend is True:
            try:
                bundle_bk = self.board.params["usb"]["bundle_mode"]
                bk_tx_pause = self.board.params["usb"]["tx_pause"]
                self.board.params["usb"]["bundle_mode"] = True
                self.board.params["usb"]["tx_pause"] = 0
            except AttributeError:
                bundle_bk = None
                bk_tx_pause = None
        else:
            try:
                bundle_bk = self.board.connection.bundle_mode
                bk_tx_pause = self.board.connection.tx_pause
                self.board.connection.bundle_mode = True
                self.board.connection.tx_pause = 0
            except AttributeError:
                bundle_bk = None
                bk_tx_pause = None
        for i, value in enumerate(scan_values):
            if self._cancel:
                break
            self._progress.append(
                (int(95 * i / len(scan_values)), f"Scanning {(i+1)}/{len(scan_values)}")
            )
            time.sleep(pause)

            scals = self._get_scaler_value(value)

            for chan in self.channels:
                output[chan][i] = scals[chan]
        self._set_ctrl_reg("trigger_select", trigger_select_backup)

        if bundle_bk is not None and bk_tx_pause is not None:
            if self.board.using_new_backend is True:
                self.board.params["usb"]["bundle_mode"] = bundle_bk
                self.board.params["usb"]["tx_pause"] = bk_tx_pause
            else:
                self.board.connection.bundle_mode = bundle_bk
                self.board.connection.tx_pause = bk_tx_pause

        return output

    def _get_scaler_value(self, trigger_value) -> Dict[int, int]:
        """Read the scalers for selected channels for a specific trigger value.

        Args:
            channels (list): list of channels to scan the scalers

        Returns:
            dict of scaler values for the selected channels
        """
        scalers = {}
        for chan in self.channels:
            scal = self._read_scaler(chan, trigger_value)
            scalers[chan] = scal

        return scalers

    def _read_scaler(self, channel, trigger_value):
        """Read the scaler value for a specific channel and trigger value.

        Args:
            channel (int): channel to read the scaler from
            trigger_value (int): trigger value to scan

        Returns:
            Scaler value for the channel
        """
        trigger_vals = [0 for _ in range(self.board.channels)]
        trigger_vals[channel] = int(trigger_value)
        tc = get_trigger_controller(self.board)
        tc.trigger_values = trigger_vals
        self._reset_counter()
        scal = self._get_scalar()
        return scal

    def _get_scalar(self) -> int:
        """Read and combine the scalar values from the two 16-bit registers."""
        upper = self._get_ctrl_reg("count_msb")
        lower = self._get_ctrl_reg("count_lsb")
        scaler = (upper << 16) + lower
        return scaler

    def _get_ctrl_reg(self, name: str):
        return ControlRegisters(self.board).read(name)["value"]

    def _set_ctrl_reg(self, name: str, value: int):
        ControlRegisters(self.board).write(name, value)

    def _reset_counter(self):
        """Reset the counter to 0."""
        self._set_ctrl_reg("count_reset", 1)
        self._set_ctrl_reg("count_reset", 0)



from typing import Callable

from bayes_opt import BayesianOptimization, UtilityFunction

from naludaq.helpers import validations
from naludaq.helpers.exceptions import OperationCanceledError


class BayesianOptimizer:
    def __init__(self, cost_function, bounds, **kwargs):
        """Wrapper for the bayes_opt module.

        Bayesian optimization is used for black-box functions where
        you want to find the maximum output by probing the input parameters, using Bayes Theorem to
        estimate the best next input parameters to try.

        This class may be used either as a one-off, or iteratively. If used iteratively, the
        cost function need not be specified, as it is not used -- the suggest and register methods
        are used instead to probe the input space.

        This constructor initializes a Bayesian Optimization object, which will try to find the maximum
        output of a cost function by probing within the given bounds. Discrete input space is not supported.
        If you want discrete inputs only, that conversion has to happen within the cost function.

        Args:
            cost_function (Callable): Function that returns a single float with inputs x and y
            bounds (dict): Dictionary with bounds of the parameter space.
                Example: {'x': [1100, 1500], 'y': [900, 1500]}

        References: https://distill.pub/2020/bayesian-optimization/
        """
        self._cancel = False
        self._progress: list = []

        self.optimizer = BayesianOptimization(f=cost_function, pbounds=bounds, **kwargs)
        self.iteration_history = self.optimizer.res
        self._cancel_flag = 0
        self.iterations = 50
        self._utility = UtilityFunction(kind="ucb", kappa=2.5, xi=0.0)

    @property
    def maximum(self) -> dict:
        """Get the current maximum output value found by the optimizer."""
        return self.optimizer.max

    def maximize(self, n_iter: int) -> dict:
        """
        Finds the input values that result in the maximum output value when run through the cost
        function.

        Args:
            n_iter (int): Number of iterations to run. (i.e. how many attempts to find max value.)

        Returns:
            dict: Dictionary with the stored input params that resulted in the largest output value.
        """
        self._utility = UtilityFunction(kind="ucb", kappa=2.5, xi=0.0)
        for i in range(n_iter):
            self._raise_if_canceled()
            next_point_to_probe = self.suggest()
            target = self.cost_function(**next_point_to_probe)
            self.register(next_point_to_probe, target)
            self._progress.append((int(100 * i / n_iter), f"Scanning {(i+1)}/{n_iter}"))

        return self.optimizer.max

    def suggest(self) -> dict:
        """Suggest the next point to probe.

        Only for use in an iterative operation.

        Returns:
            dict: Dictionary with the next input parameters to try.
        """
        return self.optimizer.suggest(self._utility)

    def register(self, point: dict, value: float):
        """Register a point and its value with the optimizer.

        Only for use in an iterative operation.

        Args:
            point (dict): Dictionary with the input parameters that resulted in the given value.
            value (float): The output value of the cost function when run with the given input
        """
        self.optimizer.register(params=point, target=value)

    def run(self):
        self.maximize(self.iterations)

    def cancel(self):
        self._cancel = True

    def _raise_if_canceled(self):
        """Raise an ``OperationCanceledError`` if the cancel flag is set."""
        if self._cancel:
            raise OperationCanceledError("Pedestals generation was canceled.")

    @property
    def progress(self):
        """Get/Set the progress message queue.

        This is a hook to read the progress if running threads.
        """
        return self._progress

    @progress.setter
    def progress(self, value):
        if not hasattr(value, "append"):
            raise TypeError(
                "Progress updates are stored in an object with an 'append' method"
            )
        self._progress = value

    @property
    def cost_function(self):
        """Get/Set the cost function.

        Can be set to None when this object is used for iterative operations.
        """
        return self._cost_function

    @cost_function.setter
    def cost_function(self, cost_func):
        if not type(Callable):
            raise TypeError("Cost Function has to be a function")
        if cost_func is not None:
            validations.validate_callable_or_raise(cost_func)
        self._cost_function = cost_func



import os

import setuptools
from setuptools import setup

CURR_DIR = os.path.abspath(os.path.dirname(__file__))


def readme():
    with open(os.path.join(CURR_DIR, "README.md"), "r", encoding="utf-8") as fh:
        return fh.read()


def getver():
    version = dict()
    with open(os.path.join(CURR_DIR, "src", "naludaq", "_version.py")) as fp:
        exec(fp.read(), version)
        return version["__version__"]


setup(
    name="naludaq",
    version=getver(),
    author="Marcus Luck",
    author_email="marcus@naluscientific.com",
    description="Backend package for Nalu hardware",
    python_requires=">=3.9",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="",
    install_requires=[
        "naluconfigs>=13.3.0",
        "ftd3xx>=1.0",
        "naludaq_rs>=0.1.10",
        "deprecation",
        "ftd2xx>=1.1.2",
        "h5py",
        "numpy<1.24",
        "pyserial==3.4",
        "pyyaml>=5.1.1",
        "scikit-learn==1.1.3",
        "requests",
    ],
    tests_require=[
        "pytest==5.0.1",
        "pytest-mock",
        "pytest-xdist",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)



import logging

from naludaq.communication import _fpga
from naludaq.helpers.exceptions import I2CError, InvalidBoardModelError
from naludaq.helpers.semiton import SemitonABC

logger = logging.getLogger("naludaq.i2c_registers")


class I2CRegisters(_fpga.FpgaRegisters, SemitonABC):
    def __init__(self, board):
        """Communication with I2C devices through FPGA registers.

        Args:
            board (Board): the board object.
        """
        if "i2c_registers" not in board.registers:
            raise InvalidBoardModelError(
                f'Board "{board.model}" is missing I2C registers or does not support I2C'
            )
        super().__init__(board, "i2c_registers")

        # These are the same on all valid boards
        self._i2c_transmit_command = board.params.get("i2c", {}).get(
            "transmit_command", "CA000000"
        )

    def transmit_command(self):
        """Send the command which tells the board to transmit data over the I2C bus.

        Raises:
            I2CError: if the command fails to send.
        """
        try:
            self._send_command(self._i2c_transmit_command)
        except Exception as e:
            raise I2CError("Failed to initiate I2C send") from e



import numpy as np

from naludaq.helpers.exceptions import BadDataError

from .parser import Parser


class Upac96Parser(Parser):
    """UPAC96 Parser

    Data Format
    For each chip:
        - Chip header (1 byte, 0x80)
        For each channel
            - 1 window header (2 bytes, 0xCCWW)
            - 4096 samples (2 bytes each: 6-bits ECC + 10 bits data)
            - 1 window footer (2 byte CRC, 0xZZFC)
        - Chip footer (3 bytes, 0x00FACE)
    """

    def __init__(self, params: dict):
        super().__init__(params)

        # physical parameters
        self._chips = params.get("num_chips", 6)
        self._channels = params.get("channels", 96)
        self._channels_per_chip = params.get(
            "channels_per_chip", self.channels // self._chips
        )
        self._windows = params.get("windows", 64)
        self._samples = params.get("samples", 64)

        # masks & shifts
        self._data_mask = params.get("data_mask", 0x03FF)  # 10 bits
        self._window_mask = params.get("windmask", 0x00FF)
        self._channel_mask = params.get("chanmask", 0xFF00)
        self._channel_shift = params.get("chanshift", 8)

        # headers & footers
        self._event_footers = params.get("event_footers", 2)  # bytes
        self._chip_header_bytes = params.get("chip_headers", 1)  # bytes
        self._chip_footer_bytes = params.get("chip_footers", 3)  # bytes
        self._window_headers = params.get("window_headers", 1)  # words
        self._window_footers = params.get("window_footers", 1)  # words

        self.invert_data = True

    @property
    def invert_data(self) -> bool:
        """Enable data inversion.

        If True, inverts each data point to (1023 - data)
        """
        return bool(self._max_data_value)

    @invert_data.setter
    def invert_data(self, val: bool):
        if val:
            # Bitwise XOR of 1's results in subtraction (inversion)
            self._max_data_value = 0x03FF
        else:
            # Bitwise XOR of 0's results in no bit flips
            self._max_data_value = 0x0000

    def parse_digital_data_old(self, in_data) -> dict:
        """Parse a UPAC96 raw event into the event dict structure.

        Args:
            in_data (dict): dict with 'rawdata' key.

        Returns:
            dict: the event dict.
        """
        self._validate_input_package(in_data)

        # cache useful values to limit lookup
        channels = self._channels
        channels_per_chip = self._channels_per_chip
        windows = self._windows
        samples = self._samples
        total_samples = windows * samples
        chip_header_bytes = self._chip_header_bytes
        chip_footer_bytes = self._chip_footer_bytes
        window_footers = self._window_footers
        window_headers = self._window_headers
        channel_shift = self._channel_shift
        channel_mask = self._channel_mask
        window_mask = self._window_mask
        data_mask = self._data_mask
        event_footers = self._event_footers

        # output goes into these arrays, faster than dict lookup each time
        output_data = np.full((channels, total_samples), np.nan)
        output_window_labels = np.full((channels, windows), np.nan)

        # Figure out how many chips there are in the data
        channel_stride = window_headers + total_samples + window_footers
        chip_stride_bytes = (
            chip_header_bytes
            + 2 * channel_stride * channels_per_chip
            + chip_footer_bytes
        )
        num_chips = (len(in_data["rawdata"]) - event_footers) // chip_stride_bytes

        # 2. Raw data gets reshaped into axes (chip, byte in chip) so we can pull out the start of each chip later
        raw_data = np.frombuffer(in_data["rawdata"], dtype=np.uint8)[:-event_footers]
        # print(raw_data)
        raw_data_matrix = raw_data.reshape((num_chips, chip_stride_bytes))

        # Data starts on odd bytes, need to strip out the beginning of each chip's data
        chip_data_matrix = np.reshape(
            np.frombuffer(
                raw_data_matrix[:, chip_header_bytes:-chip_footer_bytes].flatten(), "<H"
            ),
            (num_chips, channel_stride * channels_per_chip),
        )
        start_window = chip_data_matrix[0, 0] & window_mask

        chip_numbers = raw_data_matrix[:, 0] & 0x3F
        window_range = np.arange(windows)  # reused for window label computation
        for channel in range(channels_per_chip):
            # Grab the window headers for all chips with corresponding channel numbers at once.
            # This way we can do fancy indexing to parse more data with numpy and less with python
            current_window_headers = chip_data_matrix[:, channel * channel_stride]
            current_window_labels = current_window_headers & window_mask
            current_channels = (
                (current_window_headers & channel_mask) >> channel_shift
            ) + chip_numbers * channels_per_chip

            data_start = channel * channel_stride + window_headers
            data_stop = data_start + total_samples
            output_data[current_channels] = (
                chip_data_matrix[:, data_start:data_stop] & data_mask
            ) ^ self._max_data_value

            # Same as UDC16, start window can change but following window numbers are always in order
            window_labels = (current_window_labels[:, None] + window_range) % windows
            output_window_labels[current_channels] = window_labels

        # do some extra conversion to get the data into the right format
        output_data = self._remove_nan_channels(output_data)
        output_window_labels = self._remove_nan_channels(output_window_labels)
        output_window_labels = self._cast_sub_array_dtype(
            output_window_labels, np.uint16
        )
        start_windows = [(list(x) + [0])[0] for x in output_window_labels]

        event = {
            "chips": chip_numbers,
            "data": output_data,  # list[np.ndarray[float]]
            "window_labels": output_window_labels,  # list[np.ndarray[u16]]
            "start_window": start_window,  # int
            "start_windows": start_windows,  # list[int]
        }

        return event

    def _add_xaxis_to_event(self, event):
        """Create a time axis for the event. Time axis is computed
        based on the start window on a per-channel basis.

        Returns:
            numpy array with sample numbers for each channel.
        """
        times = []
        samples = self.samples
        for window_labels, start_window in zip(
            event["window_labels"], event["start_windows"]
        ):
            winds = np.array(window_labels, dtype="int32")
            winds += self.windows * (winds < start_window)
            winds -= start_window

            timmy = (samples * np.repeat(winds, samples)) + (
                np.ones((len(winds), samples), dtype=int)
                * np.arange(0, samples, dtype=int)
            ).flatten()
            times.append(timmy)

        return times

    def _validate_input_package(self, in_data):
        """Run preliminary checks on the input package to see if it can be parsed."""
        raw_data = in_data["rawdata"]
        data_len = len(raw_data) - self._event_footers
        instance_data_len = (
            self._chip_header_bytes
            + (
                self.windows * self.samples
                + self._window_headers
                + self._window_footers
            )
            * 2
            * self._channels_per_chip
            + self._chip_footer_bytes
        )

        if data_len & 1 != 0:
            raise BadDataError(f"Data is not an even number of bytes (got {data_len})")
        if data_len % instance_data_len != 0:
            raise BadDataError(
                f"Data length ({data_len} bytes) is not divisible "
                f"by instance data size ({instance_data_len} bytes)"
            )

    def _align_event_windows(
        self, data: list[np.ndarray], window_labels: list[np.ndarray], start_window: int
    ):
        """Align windows in the data/window label array based on the starting window. Modification is in-place."""
        samples = self._samples
        for chan, (chan_data, chan_window_labels) in enumerate(
            zip(data, window_labels)
        ):
            if len(chan_data) == 0 or len(chan_window_labels) == 0:
                continue
            window_roll_amt = chan_window_labels[0] - start_window
            data[chan] = np.roll(chan_data, window_roll_amt * samples)
            window_labels[chan] = np.roll(chan_window_labels, window_roll_amt)

    @staticmethod
    def _remove_nan_channels(arr: np.ndarray) -> list[np.ndarray]:
        """Replace nan-filled channels with an empty array.

        Args:
            arr (np.ndarray): 2D array

        Returns:
            list[np.ndarray]: list of arrays.
        """
        return [np.array([]) if np.any(np.isnan(c)) else c for c in arr]

    @staticmethod
    def _cast_sub_array_dtype(arr: list[np.ndarray], dtype) -> list[np.ndarray]:
        """Cast dtype of arrays within a list"""
        result = []
        for x in arr:
            if not isinstance(x, np.ndarray):
                x = np.array(x)
            result.append(x.astype(dtype))
        return result



"""Subpackage for software drivers used talk to devices on the board through I2C or SPI.

Examples of devices include temperature sensors and voltage monitors.
"""
from .eeprom import EEPROM
from .ltc2990 import LTC2990



"""Readout settings for the board.

Controls the readout channels and the readout windows.
"""
from logging import getLogger
from typing import List, Tuple

from naludaq.communication import AnalogRegisters, ControlRegisters, DigitalRegisters
from naludaq.controllers.controller import Controller

LOGGER = getLogger("naludaq.readout_controller_default")
MAX_READOUT = 65536


class ReadoutController(Controller):
    """Handles all logic for readout settings.


    Functions:
    set_read_window
    set_number_events_to_read
    set_readout_channels
    get_readout_channels

    """

    def __init__(self, board):
        super().__init__(board)

    @property
    def windows(self):
        """Get the number of windows to read."""
        reg = "readoutwindows"
        return self.board.registers["digital_registers"][reg]["value"][0]

    @property
    def lookback(self):
        """Get the number of windows to look back."""
        reg = "readoutlookback"
        return self.board.registers["digital_registers"][reg]["value"][0]

    @property
    def write_after_trig(self):
        """Get the number of windows to write after trigger."""
        reg = "writeaftertrig"
        return self.board.registers["digital_registers"][reg]["value"][0]

    def set_record_window(self, record_length: int, pre_trigger: int):
        """Set the read window for non-forced mode using the pre-trigger
        and record length.

        Args:
            pre_trigger (int): the number of windows before the trigger event
                to start reading at.
            record_length (int): the number of windows to read.
        """
        # validate
        if not isinstance(pre_trigger, int):
            raise TypeError(
                f"Horizontal position should be int, got {type(pre_trigger)}."
            )
        if not isinstance(record_length, int):
            raise TypeError(
                f"Horizontal scale should be int, got {type(record_length)}."
            )
        if not 0 <= pre_trigger <= self.board.params["windows"]:
            raise ValueError(f"Horizontal position out of bounds, got {pre_trigger}.")
        if not 1 <= record_length <= self.board.params["windows"]:
            raise ValueError(f"Horizontal scale out of bounds, got {record_length}.")

        # Logic
        window, lookback, write_after_trig = self.record_to_readout_window(
            record_length, pre_trigger
        )
        return self.set_read_window(
            windows=window,
            lookback=lookback,
            write_after_trig=write_after_trig,
        )

    def record_to_readout_window(
        self, record_length: int, pre_trigger: int
    ) -> Tuple[int, int, int]:
        """Get the read window for non-forced mode using the pre-trigger and record length."""
        wat_min = 0
        write_after_trig = max(record_length - pre_trigger, wat_min)
        lookback = pre_trigger if write_after_trig == wat_min else record_length
        windows = record_length
        return (windows, lookback, write_after_trig)

    def record_to_readout_window_forced(
        self, record_length: int, pre_trigger: int
    ) -> Tuple[int, int, int]:
        """Get the read window for non-forced mode using the pre-trigger and record length."""
        write_after_trig = self.board.params["windows"]
        lookback = pre_trigger
        windows = record_length
        return (windows, lookback, write_after_trig)

    def read_to_record_window(
        self, windows: int, lookback: int, write_after_trig: int
    ) -> Tuple[int, int]:
        """Get the read window for non-forced mode using the pre-trigger and record length.

        This function is missing the scenario where the write_after trigger is bigger than lookback.
        This is because the NaluScope logic doesn't allow for this scenario.

        Currently it will return the pre_trigger as negative.
        """
        record_length = windows
        pre_trigger = max(lookback - write_after_trig, 0)
        return (record_length, pre_trigger)

    def read_to_record_window_forced(
        self, windows: int, lookback: int, write_after_trig: int = None
    ) -> Tuple[int, int]:
        """Get the read window for forced mode using the start window and record length."""
        record_length = windows
        start_window = lookback
        return (record_length, start_window)

    def set_record_window_forced(self, record_length: int, start_window: int):
        """Set the read window for forced mode using the start window
        and record length.

        Args:
            start_window (int): the window in the sampling array to start at.
            record_length (int): the number of windows to read.
        """
        return self.set_read_window(
            windows=record_length,
            lookback=start_window,
            write_after_trig=record_length,
        )

    def set_read_window(self, windows=None, lookback=None, write_after_trig=None):
        """Setup the readwindow on the timescale.

        Args:
            windows(int): Amount of readout windows to read.
            lookback(int): How many windows to lookback after trigger, or in forced mode it's the
                actual window number.
            write_after_trigger(int): readout addresses to write after trig. ! 1 addr = 2 windows!
        """
        if windows is not None:
            if not isinstance(windows, int):
                raise TypeError(f"windows should be int or None, got {type(windows)}")
            if not 0 <= windows <= self.board.params["windows"]:
                raise ValueError(f"windows out of bounds, got {windows}")

        if lookback is not None:
            if not isinstance(lookback, (type(None), int)):
                raise TypeError(f"lookback should be int or None, got {type(lookback)}")
            if not 0 <= lookback <= self.board.params["windows"]:
                raise ValueError(f"lookback out of bounds, got {lookback}")

        if write_after_trig is not None:
            # self.max_write_after_trig
            wat_max_bit = None
            wat_max_bit = (
                self.board.params.get("registers", {})
                .get("digital_registers", {})
                .get("writeaftertrig", {})
                .get("bitwidth", 12)
            )
            if not wat_max_bit:
                wat_max_bit = 12
            wat_max = 2**wat_max_bit

            if not isinstance(write_after_trig, (type(None), int)):
                raise TypeError(
                    f"write_after_trig should be int or None, "
                    f"got {type(write_after_trig)}"
                )
            if not 0 <= write_after_trig <= wat_max:
                raise ValueError(
                    f"write_after_trig out of bounds, got {write_after_trig}"
                )

        LOGGER.debug(
            "Set readwindow to: w%s, l%s, t:%s", windows, lookback, write_after_trig
        )
        if self.board.params["model"].lower() == "siread":
            win_reg = "numwinds"
            lookback_reg = "lookback"
        else:
            win_reg = "readoutwindows"
            lookback_reg = "readoutlookback"
        write_after_trig_reg = "writeaftertrig"

        if windows is not None:
            self._write_read_window(win_reg, windows)
        if lookback is not None:
            self._write_read_window(lookback_reg, lookback)
        if write_after_trig is not None:
            write_after_trig = self._compute_write_after_trig(
                write_after_trig
            )  # Remember, readout address is 2 windows.
            self._write_read_window(write_after_trig_reg, write_after_trig)
        return (windows, lookback, write_after_trig)

    def _compute_write_after_trig(self, write_after_trig):
        """Most ASICs use 2 windows per readout address, so we need to divide by 2."""
        return write_after_trig // 2

    def _write_read_window(self, digital_register, value):
        """write read window settings to board.

        This is an adapter function since SiREAD use a different register.
        It will update the ocrrect register depending on the board used.

        Args:
            digital_register(str): register to write
            value(int): value to set.
        """
        if self.board.params["model"].lower() == "siread":
            self._write_control_register(digital_register, value)
        else:
            self._write_digital_register(digital_register, value)

    def number_events_to_read(self, amount: int) -> None:
        """Tell the board the maximum number of events to readout.

        This only works if the readout is set with singleEv is set to True.

        Args:
            amount (int): MAximum number of events to read. 16-bit register.
        """
        if not isinstance(amount, int):
            raise TypeError(f"amount must be an integer, got {type(amount)}")
        if not 0 <= amount < MAX_READOUT:
            raise ValueError("amount must be an positive integer smaller than 65536.")
        self._write_control_register("runevs", amount)

    def set_readout_channels(self, channels_to_read: list):
        """Select channels to readout.

        Update the registers and write them to the board.

        Args:
            channels_to_read(list): List of channel numbers to read
        """
        self._validate_channels_or_raise(channels_to_read)
        chansel = self._generate_channels_bits(channels_to_read)

        if self.board.params["model"].lower() == "siread":
            self._set_readout_channels_siread(chansel)
        else:
            self._set_readout_channels_generic(chansel)

    def _validate_channels_or_raise(self, channels: List[int], min_valid=1):
        """Validates a list of channels.

        Args:
            channels (List[int]): the list of channels
            min_valid (int): minimum number of channels that must
                be present in the list to be considered valid.

        Raises:
            TypeError if the list or an element in the list are the wrong type
            ValueError if the list is too large, too small, or contains a channel
                number that is out of bounds.
        """
        if not isinstance(channels, list):
            raise TypeError(f"Channels must be a list, got {type(channels)}.")
        if not (min_valid <= len(channels) <= self.board.channels):
            raise ValueError(
                f"List of channels needs to contain between {min_valid} and "
                f"{self.board.channels} channels."
            )
        if not all(isinstance(x, int) for x in channels):
            raise TypeError(f"Channel numbers must be integers.")
        if max(channels or [0]) > self.board.channels or min(channels or [0]) < 0:
            raise ValueError(f"At least one channel number is out of bounds.")

    def _generate_channels_bits(self, channels_to_read):
        """Generates the chansel bitstring from a list fo channels.

        Args:
            channels_to_read(list): List of channel numbers to read.

        Returns:
            String of '1' and '0' where position repsresents channel.
        """
        chansel = str()
        for i in range(self.board.params["channels"]):
            if i in channels_to_read:
                chansel = "1" + chansel
            else:
                chansel = "0" + chansel
        return chansel

    def _set_readout_channels_generic(self, chansel):
        """Select readout channels for all non-siread boards."""
        self._write_control_register("chansel", int(chansel, 2))

    def _set_readout_channels_siread(self, chansel):
        """Select readout channels for the siread board.
        The siread uses two 16 bit registers for the channels (32 ch -> 32 bit).
        """
        self._write_control_register("chanselb", int(chansel[0:16], 2))
        self._write_control_register("chansela", int(chansel[16:], 2))

    def get_readout_channels(self):
        """Get the current channels to read out.

        Returns:
            Sorted list of channel numbers to read out.
        """
        if self.board.params["model"].lower() == "siread":
            return self._get_readout_channels_siread()
        else:
            return self._get_readout_channels_generic()

    def _get_readout_channels_siread(self):
        """Returns the channels for siread since it uses two registers instead of one.

        Returns:
            List of channels currently set to read out in numerical order.
        """
        chansela = self.board.registers["control_registers"]["chansela"]
        chanselb = self.board.registers["control_registers"]["chanselb"]
        result = list()
        result.extend(
            [
                (x)
                for x, val in enumerate(reversed(list(bin(chanselb[0])[2:])))
                if int(val) == 1
            ]
        )
        result.extend(
            [
                (x)
                for x, val in enumerate(reversed(list(bin(chansela[0])[2:])))
                if int(val) == 1
            ]
        )
        return result

    def _get_readout_channels_generic(self):
        chansel = self.board.registers["control_registers"]["chansel"]
        return [
            (x)
            for x, val in enumerate(reversed(list(bin(chansel["value"])[2:])))
            if int(val) == 1
        ]

    def _write_analog_register(
        self,
        register: str,
        value: int,
        chips: "int | list[int] | None" = None,
    ):
        """Helper function for writing analog registers.

        Args:
            register (str): register name
            value (int): value to write.
        """
        AnalogRegisters(self.board, chips).write(register, value)

    def _write_control_register(self, register: str, value: int):
        """Helper function for writing control registers.

        Args:
            register (str): register name
            value (int): value to write.
        """
        ControlRegisters(self.board).write(register, value)

    def _write_digital_register(self, register: str, value: int):
        """Helper function for writing digital registers.

        Args:
            register (str): register name
            value (int): value to write.
        """
        DigitalRegisters(self.board).write(register, value)



"""Timing calibration

Uses an array of 128 timing constants to correct the naluscopes timing.
The timing constants can be generated using a separate tool.

The timing constanst will correct the spacing between the samples.
The boards clock is broken down internally in the ASIC into 128 pieces, the
general asusmtion is that the clock feeding the chip can be considered exact since it feeds all
 parts equally. Thus the 128 spacing.

The timing calibration is a tool that corrects the ASICs internal timing jidder.
The internal timing jidder is aslo assumed to be constants over time, meaning it will not change
significant from one capture to the next. Once the timing constants have been generated they can
be saved and used indefinitely.
"""
from .calibration import TimingCalibration
from .correcter import TimingCorrecter



import logging

import numpy as np

from naludaq.helpers import FancyIterator
from naludaq.helpers.exceptions import (
    IterationError,
    OperationCanceledError,
    PedestalsDataCaptureError,
)

from .default import PedestalsGenerator

LOGGER = logging.getLogger("naludaq.pedestals_generator_udc16")


class PedestalsGeneratorUdc16(PedestalsGenerator):
    """Pedestals generator for UDC16."""

    def _capture_data_for_pedestals(self) -> list[list[dict]]:
        """Capture raw data for pedestals.

        For UDC16/UPAC96 it's a bit different. We need to throw away the first window and there's no forced mode,
        so it's random where the window hole shows up in the data. This capture function makes sure that
        regardless of where the hole shows up in each event, we have enough captures for all windows.

        Returns:
            list[list[dict]]: list of data for blocks. Warmup events
                are removed from the output.

        Raises:
            PedestalsDataCaptureError: if pedestals failed to generate.
            OperationCanceledError: if pedestals generation was cancelled.
        """
        LOGGER.debug(
            "Capturing pedestals for %s. channels=%s",
            self.board.model,
            self._channels,
        )
        get_actual_count = lambda events: np.min(
            self._get_window_counts(events)[self.channels]
        )
        pipeline = (
            self._create_data_pipeline()
            .accumulated()
            .take_while(lambda all: get_actual_count(all) <= self._num_captures)
            .unaccumulated()
        )
        try:
            return [pipeline.collect()]
        except OperationCanceledError:
            raise
        except (TimeoutError, IterationError) as e:
            LOGGER.error("Failed to generate pedestals data! %s", exc_info=e)
            msg = self._exc_str(e)
            raise PedestalsDataCaptureError(msg) from e

    def _create_validation_stage(self, pipeline: FancyIterator) -> FancyIterator:
        return pipeline.filter(self._validate_event, exclusion_limit=10)

    def _validate_event(self, event: dict) -> bool:
        """Validate an event.

        The expected_window_labels argument is ignored for UDC16,
        since if it's parsed then it's probably valid.
        """
        return "data" in event

    def _backup_settings(self) -> dict:
        return {}

    def _create_progress_update_stage(self, pipeline: FancyIterator) -> FancyIterator:
        n_warmup = self.num_warmup_events
        total_events = n_warmup + self._num_captures
        min_progress = 20
        max_progress = 80

        def inner(events):
            idx = len(events)
            if idx > n_warmup:
                window_counts = self._get_window_counts(events[n_warmup:])
                n_actual_events = np.min(window_counts[self.channels])
                idx = n_warmup + n_actual_events
            idx -= 1
            percent = min_progress + (max_progress - min_progress) * idx / total_events
            msg = f"Capturing event {idx}/{total_events}"
            self._update_progress(percent, msg)

        return pipeline.accumulated().for_each(inner).unaccumulated()

    def _store_raw_data(self, blocks: list[list[dict]]):
        num_captures = self._num_captures
        channels = self.board.params.get("channels", 16)
        windows = self.board.params.get("windows", 64)
        samples = self.board.params.get("samples", 64)
        rawdata = self.board.pedestals["rawdata"]
        window_counts = np.zeros((channels, windows), dtype=int)
        for event in blocks[0]:  # only one block for UDC16/UPAC96
            for chan in range(channels):
                chan_data = event["data"][chan]
                for window_idx, window in enumerate(event["window_labels"][chan]):
                    # first/last windows are junk
                    if window_idx == 0 or window_idx == windows - 1:
                        continue

                    cap = window_counts[chan, window]
                    window_counts[chan, window] += 1
                    if cap >= num_captures:  # only take the first "num_captures" events
                        continue

                    data = chan_data[window_idx * samples : (window_idx + 1) * samples]
                    rawdata[chan, window, :, cap] = data

        return True

    def _get_window_counts(self, events: list[dict]) -> np.ndarray:
        """Calculate the number of times each window occurs in the given events.

        Args:
            events (list[dict]): list of validated events.

        Returns:
            np.ndarray: 2D int array with shape (channels, windows) containin
                window counts.
        """
        channels = self.board.channels
        windows = self.board.params["windows"]
        window_hits = np.zeros((channels, windows), dtype=int)

        # Count the number of times each window shows up in the data.
        for event in events:
            for chan, chan_window_labels in enumerate(event["window_labels"]):
                if len(chan_window_labels) == 0:
                    continue
                # skip first/last windows, they're junk.
                window_hits[chan][chan_window_labels[1:-1]] += 1
        return window_hits

    def _exc_str(self, e: Exception) -> str:
        """Get a string describing pedestals generation errors/possible fixes.

        Two types of exceptions are relevant here:
        - TimeoutError: when no valid data is received for a while.
        - IterationError: when we get valid data but it's filtered out by the
            constraints in `validate_event()` or somewhere else.

        Args:
            e (Exception): exception that was raised.
        """
        return (
            "Failed to generate pedestals calibration. Please try again and power cycle the board "
            "if the problem persists."
        )



"""Tool for waiting for events to fill a buffer to a specified amount

Author:
    Alvin Yang
    Mitchell Matsumori-Kelly <mitchell@naluscientific.com>
"""
import logging
import time
from typing import Callable

from naludaq.tools.autoaction import AutoAction

logger = logging.getLogger("naludaq.tools.eventwaiter")


class EventWaiter(AutoAction):
    def __init__(
        self,
        buffer,
        amount,
        interval: "float | Callable" = 0.1,
        timeout: float = 1.0,
        reset_timeout=False,
    ) -> None:
        """Module for waiting for when a buffer (events) is filled to a specified amount

        Args:
            buffer: A list like object that stores events
            amount (int): Desired amount of events for the buffer to be filled, before waiter stops
            interval (float): Time (s) between each buffer check
            timeout (float): Time (s) before the waiter stops
            reset_timeout (bool): Whether to reset timeout counter when len of buffer is changed
        """
        self.buffer = buffer
        self.amount = amount
        self.timeout = timeout
        self.reset_timeout = reset_timeout
        super().__init__(self._actioner, self._stopper, interval=interval)

    def _actioner(self):
        if len(self.buffer) > self._curr_len and self.reset_timeout:
            self._curr_len = len(self.buffer)
            self._start_time = time.time()

    def _stopper(self):
        if len(self.buffer) >= self.amount:
            return True
        time_passed = time.time() - self._start_time
        if time_passed > self.timeout:
            raise TimeoutError("Waited too long for an event")
        return False

    def start(self, blocking=True):
        """Starts calling the action periodically from a new thread.

        Args:
            blocking (bool): whether to block until the stop condition is met
                or the `stop()` method is called.
        """
        self._start_time = time.time()
        self._curr_len = len(self.buffer)
        super().start(blocking=blocking)



"""
"""
from logging import getLogger

from naludaq.backend.managers.connection import ConnectionManager
from naludaq.backend.models.device import DeviceType
from naludaq.communication import ControlRegisters
from naludaq.controllers.controller import Controller

LOGGER = getLogger("naludaq.board_controller_upac")
NON_UDP = [DeviceType.SERIAL, DeviceType.D2XX, DeviceType.D3XX]

TRIGGERS = {
    "software": "00",
    "auto": "01",
    "external": "10",
    "self": "11",
}


class UpacBoardController(Controller):
    """The board control communicates with to the board.

    This class is stateless, you need to supply the board to the class to perform actions on it.
    There is no need to store the controller, it will only be used to control the board.

    It is the interface between the application and the board.
    The Analytics package or the GUI uses this class to communicate with the
    hardware. Keeps track of all registers etc.
    It doesn't receive continous data, the data acquisition module should be used for that.
    """

    def __init__(self, board):
        super().__init__(board)

    def start_readout(
        self,
        trigger_type: str = "software",
    ):
        """Sends a readout signal to the board.

        Obstensibly reads out the chip?

        Args:
            trigger_type (str):
                "There are 4 trigger modes:

                ‘00’	: 	Software Trigger Mode, the power on default.
                ‘01’	: 	Auto Trigger Mode. 1-2Hz auto trigger data rate.
                ‘10’	:	External Trigger Mode.
                ‘11’	: 	Self Trigger Mode.

        Command
        -------

        command is: BMMXXVVV
        MM = singleEvent(7) + readoutEn(6) + pedSel(5 dt 4) + modeSel(3 dt 0)
        - modeSel=x"0" analog write
        - modeSel=x"1" digital write
        - modeSel=x"2" digital read
        - modeSel=x"3" waveform read

        VVVVV = addr_sel(19 dt 12) + data_sel(11 dt 0)
        - data_sel(5 downto 4) = trig type
        - data_sel(3 downto 2) = lb type
        - data_sel(1 downto 0) = acq type


        """
        if not isinstance(trigger_type, str):
            raise TypeError(f"trigger_type must be a str, got {type(trigger_type)}")
        if trigger_type.lower() in TRIGGERS.values():
            trigger_mode = trigger_type
        elif trigger_type.lower() in TRIGGERS.keys():
            trigger_mode = TRIGGERS[trigger_type]
        else:
            raise ValueError(
                f"trig not valid, got {trigger_type}, "
                "options are 'software', auto', 'external', 'self'"
            )
        self._write_control_register("trigger_mode", int(trigger_mode, 2))

        # Needed to enable self trigger
        self._write_control_register("self_trigger_or_enable", trigger_mode == "11")

    def stop_readout(self):
        """Toggles the "stopacq" signal on the readout module.

        It's equivalent of asking it nicely to stop reading.
        """
        # if self.trigger_mode == "00":
        #     self._write_control_register("software_trigger", False)
        self._write_control_register("trigger_mode", int("00", 2))

    def toggle_trigger(self):
        """Toggles the software trigger bit, should result in a software trigger event."""
        LOGGER.debug("Toggle software trigger")
        self._write_control_register("software_trigger", True)
        self._write_control_register("software_trigger", False)

    def toggle_reread(self):
        """"""
        self._write_control_register("reread_data", True)
        self._write_control_register("reread_data", False)

    def reset_board(self):
        """Try and reset the board.

        In case the FPGA get stuck this can help reset the state.
        """
        # time.sleep(0.02)
        # self.digital_reset()
        # time.sleep(0.02)
        # self.sysrst()

    def digital_reset(self):
        """Toggles the "reset" port on the readout module.

        Forcibly returns the chip to default state.
        """
        # self._write_control_register("digrst", True)
        # self._write_control_register("digrst", False)

    def clear_buffer(self):
        """Clears the UART buffer on both CPU and FPGA side."""
        try:
            self._clear_fpga_buffer()
        except Exception:
            pass

        try:
            self._clear_uart_buffer()
        except Exception:
            pass

    def _clear_fpga_buffer(self):
        self._write_control_register("wave_fifo_reset", True)
        self._write_control_register("wave_fifo_reset", False)

    def _clear_uart_buffer(self):
        if self.board.using_new_backend:
            device = ConnectionManager(self.board).device
            if device.type in NON_UDP:
                device.clear_buffers()
        else:
            self.board.connection.reset_input_buffer()

    def sysrst(self):
        """Toggles the sysrst pin, which resets the digital portion of the chip"""
        # self._write_control_register("sysrst", True)
        # self._write_control_register("sysrst", False)

    #############################################################################
    # Random Stuff, may or maynot work... Below functions need to be validated.
    #############################################################################

    def read_scalers(self, channels: list[int] = None):
        """Not implemented."""
        # self.stop_readout()
        scalers = []
        # for i in range(0, 4):
        #     name = "scal" + str(i)
        #     scalers.append(DigitalRegisters(self.board).read(name))
        #     # if Print: logger.debug("%s %s", name, scalers[-1])

        # scalers.append(DigitalRegisters(self.board).read("scalmon"))
        # if Print: logger.debug("scalmon %s", scalers[-1])
        return scalers

    def read_firmware_version(self):
        """Read the firmware version.

        This is a control register.
        """
        result = 0
        try:
            result = ControlRegisters(self.board).read("version_number")
        except:
            LOGGER.error("Can't read firmware version")
        return result

    def read_identifier(self) -> str:
        """Read the board identifier."""
        result = ControlRegisters(self.board).read("identifier")
        return result

    def enable_testmode(self, enabled: bool):
        """Enables the test-pattern output.

        Set the fpga in a mode to output a known test-pattern.
        """
        # ControlRegisters(self.board).write("fake", enabled)

    def get_available_chips(self) -> list[int]:
        """Get a list of available chip numbers."""
        return list(range(self.board.params.get("num_chips", 1)))

    def set_lookahead(self):
        """Set look ahead mode"""
        # Old values
        name = "xref_address_mode"
        value = False
        self._write_control_register(name, value)

    def set_lookback(self):
        """Set look back mode"""
        name = "xref_address_mode"
        value = True
        self._write_control_register(name, value)

    def set_fast_rate(self):
        """Set sampling rate to 9.9Gsps.

        •"AF0C8050" Write Reg 12 x8050
        •"AF0D5777" Write Reg 13 x5777
        •"AF0EEO21" Write Reg 14 xE021
        •"AFOF8383" Write Reg 15 x8383
        •"AC130800" Set Reg 19 bit 11
        •"AB130800" Clear Reg 19 bit 11

        9.2Gsps (70Mhz)
        REG 0Chex cdce62002_register_1_msw_out 1104
        REG 0Dhex cdce62002_register_0_msw_out 21840
        REG OEhex cdce62002_register_1_lsw_out 41105
        REG OFhex cdce62002_register_1_msw_out 33677
        """
        # Old values
        reg_writes = {
            "cdce62002_register_0_lsw_out": 0x0450,
            "cdce62002_register_0_msw_out": 0x5550,
            "cdce62002_register_1_lsw_out": 0xA091,
            "cdce62002_register_1_msw_out": 0x838D,
        }

        for key, value in reg_writes.items():
            self._write_control_register(key, value)

        self._write_strobe()

    def set_slow_rate(self):
        """Set sampling rate to 3.0GSps (23.0MHz).

        • "AF0C1050" Write Reg 12 x1050
        • "AF0D5758" Write Reg 13 x5758
        • "AF0E8011" Write Reg 14 x8011
        • "AFOF8387" Write Reg 15 x8387
        • "AC130800" Set Reg 19 bit 11
        • "AB130800" Clear Reg 19 bit 11

        CHANGE TO
        3.0Gsps (23Mhz)
        REG 0Chex cdce62002_register_0_lsw_out 1104
        REG 0Dhex cdce62002_register_0_msw_out 22248
        REG OEhex cdce62002_register_1_lsw_out 40993
        REG OFhex cdce62002_register_1_msw_out 33666
        """
        reg_writes = {
            "cdce62002_register_0_lsw_out": 0x0450,
            "cdce62002_register_0_msw_out": 0x56E8,
            "cdce62002_register_1_lsw_out": 0xA021,
            "cdce62002_register_1_msw_out": 0x8382,
        }

        for key, value in reg_writes.items():
            self._write_control_register(key, value)

        self._write_strobe()

    def _write_strobe(self):
        """Write the changes and powercycles."""
        self._write_control_register("cdce62002_write_strobe", True)
        self._write_control_register("cdce62002_write_strobe", False)
        # This is normally true since it's _n and power should be up.
        self._write_control_register("cdce62002_power_down_n", False)
        self._write_control_register("cdce62002_power_down_n", True)

    def set_pd_bias_enabled(self, enabled: bool):
        """Set whether photodiode bias is enabled"""
        if enabled not in [0, 1]:
            raise TypeError("Enable flag must be a bool")
        self._write_control_register("pd_bias_en", enabled)

    def get_pd_bias_enabled(self) -> bool:
        """Get whether photodiode bias is enabled"""
        return bool(self.board.registers["control_registers"]["pd_bias_en"]["value"])

    def _write_control_register(self, register, value):
        """wrapper for the Control register coms module.

        Args:
            register (str): name of the register to update.
            value: The register value to set.

        Raises:
            ValueError, TypeError
        """
        ControlRegisters(self.board).write(register, value)  # pragma: no cover



from naludaq.board.connections.base_serial import BaseSerialConnection
from naludaq.daq.debugdaq import DebugDaq
from naludaq.daq.debugdaq import DebugDaq as BenDaq
from naludaq.daq.lightdaq import LightDaq


def get_daq(board, *args, parsed: bool = False, debug: bool = False, **kwargs):
    """Return an instantiated Daq object based on model and preferences.

    Args:
        board: board object
        parsed: should returndata be parsed or raw?
        debug: True to use a daq with more telemetry and storage options

    Returns:
        Instantiated daq object
    """
    daq = LightDaq
    if debug or parsed:
        daq = DebugDaq

    return daq(board, *args, **kwargs)
